# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Iterator,
    Optional,
    Sequence,
)
import math
import numpy as np

from ezdxf.math import (
    Vec2,
    Vec3,
    UVec,
    Z_AXIS,
    OCS,
    UCS,
    Matrix44,
    BoundingBox,
    ConstructionEllipse,
    cubic_bezier_from_ellipse,
    Bezier4P,
    Bezier3P,
    BSpline,
    reverse_bezier_curves,
    bulge_to_arc,
    linear_vertex_spacing,
    inscribe_circle_tangent_length,
    cubic_bezier_arc_parameters,
    cubic_bezier_bbox,
    quadratic_bezier_bbox,
)
from ezdxf.math.triangulation import mapbox_earcut_2d
from ezdxf.query import EntityQuery

from .path import Path
from .commands import Command
from . import converter, nesting

if TYPE_CHECKING:
    from ezdxf.query import EntityQuery
    from ezdxf.eztypes import GenericLayoutType


__all__ = [
    "bbox",
    "precise_bbox",
    "fit_paths_into_box",
    "transform_paths",
    "transform_paths_to_ocs",
    "render_lwpolylines",
    "render_polylines2d",
    "render_polylines3d",
    "render_lines",
    "render_hatches",
    "render_mpolygons",
    "render_splines_and_polylines",
    "add_bezier4p",
    "add_bezier3p",
    "add_ellipse",
    "add_2d_polyline",
    "add_spline",
    "to_multi_path",
    "single_paths",
    "have_close_control_vertices",
    "lines_to_curve3",
    "lines_to_curve4",
    "fillet",
    "polygonal_fillet",
    "chamfer",
    "chamfer2",
    "triangulate",
    "is_rectangular",
]

MAX_DISTANCE = 0.01
MIN_SEGMENTS = 4
G1_TOL = 1e-4
IS_CLOSE_TOL = 1e-10


def to_multi_path(paths: Iterable[Path]) -> Path:
    """Returns a multi-path object from all given paths and their sub-paths.
    Ignores paths without any commands (empty paths).
    """
    multi_path = Path()
    for p in paths:
        multi_path.extend_multi_path(p)
    return multi_path


def single_paths(paths: Iterable[Path]) -> Iterable[Path]:
    """Yields all given paths and their sub-paths as single path objects."""
    for p in paths:
        if p.has_sub_paths:
            yield from p.sub_paths()
        else:
            yield p


def transform_paths(paths: Iterable[Path], m: Matrix44) -> list[Path]:
    """Transform multiple path objects at once by transformation
    matrix `m`. Returns a list of the transformed path objects.

    Args:
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        m: transformation matrix of type :class:`~ezdxf.math.Matrix44`

    """
    return [p.transform(m) for p in paths]


def transform_paths_to_ocs(paths: Iterable[Path], ocs: OCS) -> list[Path]:
    """Transform multiple :class:`Path` objects at once from WCS to OCS.
    Returns a list of the transformed :class:`Path` objects.

    Args:
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        ocs: OCS transformation of type :class:`~ezdxf.math.OCS`

    """
    t = ocs.matrix.copy()
    t.transpose()
    return transform_paths(paths, t)


def bbox(paths: Iterable[Path], *, fast=False) -> BoundingBox:
    """Returns the :class:`~ezdxf.math.BoundingBox` for the given paths.

    Args:
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        fast: calculates the precise bounding box of Bèzier curves if
            ``False``, otherwise uses the control points of Bézier curves to
            determine their bounding box.

    """
    box = BoundingBox()
    for p in paths:
        if fast:
            box.extend(p.control_vertices())
        else:
            bb = precise_bbox(p)
            if bb.has_data:
                box.extend((bb.extmin, bb.extmax))
    return box


def precise_bbox(path: Path) -> BoundingBox:
    """Returns the precise :class:`~ezdxf.math.BoundingBox` for the given paths."""
    if len(path) == 0:  # empty path
        return BoundingBox()
    start = path.start
    points: list[Vec3] = [start]
    for cmd in path.commands():
        if cmd.type == Command.LINE_TO:
            points.append(cmd.end)
        elif cmd.type == Command.CURVE4_TO:
            bb = cubic_bezier_bbox(
                Bezier4P((start, cmd.ctrl1, cmd.ctrl2, cmd.end))  # type: ignore
            )
            points.append(bb.extmin)
            points.append(bb.extmax)
        elif cmd.type == Command.CURVE3_TO:
            bb = quadratic_bezier_bbox(Bezier3P((start, cmd.ctrl, cmd.end)))  # type: ignore
            points.append(bb.extmin)
            points.append(bb.extmax)
        elif cmd.type == Command.MOVE_TO:
            points.append(cmd.end)
        start = cmd.end

    return BoundingBox(points)


def fit_paths_into_box(
    paths: Iterable[Path],
    size: tuple[float, float, float],
    uniform: bool = True,
    source_box: Optional[BoundingBox] = None,
) -> list[Path]:
    """Scale the given `paths` to fit into a box of the given `size`,
    so that all path vertices are inside these borders.
    If `source_box` is ``None`` the default source bounding box is calculated
    from the control points of the `paths`.

    `Note:` if the target size has a z-size of 0, the `paths` are
    projected into the xy-plane, same is true for the x-size, projects into
    the yz-plane and the y-size, projects into and xz-plane.

    Args:
        paths: iterable of :class:`~ezdxf.path.Path` objects
        size: target box size as tuple of x-, y- and z-size values
        uniform: ``True`` for uniform scaling
        source_box: pass precalculated source bounding box, or ``None`` to
            calculate the default source bounding box from the control vertices

    """
    paths = list(paths)
    if len(paths) == 0:
        return paths
    if source_box is None:
        current_box = bbox(paths, fast=True)
    else:
        current_box = source_box
    if not current_box.has_data or current_box.size == (0, 0, 0):
        return paths
    target_size = Vec3(size)
    if target_size == (0, 0, 0) or min(target_size) < 0:
        raise ValueError("invalid target size")

    if uniform:
        sx, sy, sz = _get_uniform_scaling(current_box.size, target_size)
    else:
        sx, sy, sz = _get_non_uniform_scaling(current_box.size, target_size)
    m = Matrix44.scale(sx, sy, sz)
    return transform_paths(paths, m)


def _get_uniform_scaling(current_size: Vec3, target_size: Vec3):
    TOL = 1e-6
    scale_x = math.inf
    if current_size.x > TOL and target_size.x > TOL:
        scale_x = target_size.x / current_size.x
    scale_y = math.inf
    if current_size.y > TOL and target_size.y > TOL:
        scale_y = target_size.y / current_size.y
    scale_z = math.inf
    if current_size.z > TOL and target_size.z > TOL:
        scale_z = target_size.z / current_size.z

    uniform_scale = min(scale_x, scale_y, scale_z)
    if uniform_scale is math.inf:
        raise ArithmeticError("internal error")
    scale_x = uniform_scale if target_size.x > TOL else 0
    scale_y = uniform_scale if target_size.y > TOL else 0
    scale_z = uniform_scale if target_size.z > TOL else 0
    return scale_x, scale_y, scale_z


def _get_non_uniform_scaling(current_size: Vec3, target_size: Vec3):
    TOL = 1e-6
    scale_x = 1.0
    if current_size.x > TOL:
        scale_x = target_size.x / current_size.x
    scale_y = 1.0
    if current_size.y > TOL:
        scale_y = target_size.y / current_size.y
    scale_z = 1.0
    if current_size.z > TOL:
        scale_z = target_size.z / current_size.z
    return scale_x, scale_y, scale_z


# Path to entity converter and render utilities:


def render_lwpolylines(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    distance: float = MAX_DISTANCE,
    segments: int = MIN_SEGMENTS,
    extrusion: UVec = Z_AXIS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as
    :class:`~ezdxf.entities.LWPolyline` entities.
    The `extrusion` vector is applied to all paths, all vertices are projected
    onto the plane normal to this extrusion vector. The default extrusion vector
    is the WCS z-axis. The plane elevation is the distance from the WCS origin
    to the start point of the first path.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve
        extrusion: extrusion vector for all paths
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    lwpolylines = list(
        converter.to_lwpolylines(
            paths,
            distance=distance,
            segments=segments,
            extrusion=extrusion,
            dxfattribs=dxfattribs,
        )
    )
    for lwpolyline in lwpolylines:
        layout.add_entity(lwpolyline)
    return EntityQuery(lwpolylines)


def render_polylines2d(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    distance: float = 0.01,
    segments: int = 4,
    extrusion: UVec = Z_AXIS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as 2D
    :class:`~ezdxf.entities.Polyline` entities.
    The `extrusion` vector is applied to all paths, all vertices are projected
    onto the plane normal to this extrusion vector.The default extrusion vector
    is the WCS z-axis. The plane elevation is the distance from the WCS origin
    to the start point of the first path.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve
        extrusion: extrusion vector for all paths
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    polylines2d = list(
        converter.to_polylines2d(
            paths,
            distance=distance,
            segments=segments,
            extrusion=extrusion,
            dxfattribs=dxfattribs,
        )
    )
    for polyline2d in polylines2d:
        layout.add_entity(polyline2d)
    return EntityQuery(polylines2d)


def render_hatches(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    edge_path: bool = True,
    distance: float = MAX_DISTANCE,
    segments: int = MIN_SEGMENTS,
    g1_tol: float = G1_TOL,
    extrusion: UVec = Z_AXIS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as
    :class:`~ezdxf.entities.Hatch` entities.
    The `extrusion` vector is applied to all paths, all vertices are projected
    onto the plane normal to this extrusion vector. The default extrusion vector
    is the WCS z-axis. The plane elevation is the distance from the WCS origin
    to the start point of the first path.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path` or :class:`Path2d`  objects
        edge_path: ``True`` for edge paths build of LINE and SPLINE edges,
            ``False`` for only LWPOLYLINE paths as boundary paths
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve to flatten polyline paths
        g1_tol: tolerance for G1 continuity check to separate SPLINE edges
        extrusion: extrusion vector for all paths
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    hatches = list(
        converter.to_hatches(
            paths,
            edge_path=edge_path,
            distance=distance,
            segments=segments,
            g1_tol=g1_tol,
            extrusion=extrusion,
            dxfattribs=dxfattribs,
        )
    )
    for hatch in hatches:
        layout.add_entity(hatch)
    return EntityQuery(hatches)


def render_mpolygons(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    distance: float = MAX_DISTANCE,
    segments: int = MIN_SEGMENTS,
    extrusion: UVec = Z_AXIS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as
    :class:`~ezdxf.entities.MPolygon` entities. The MPOLYGON entity supports
    only polyline boundary paths. All curves will be approximated.

    The `extrusion` vector is applied to all paths, all vertices are projected
    onto the plane normal to this extrusion vector. The default extrusion vector
    is the WCS z-axis. The plane elevation is the distance from the WCS origin
    to the start point of the first path.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path` or :class:`Path2d` objects
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve to flatten polyline paths
        extrusion: extrusion vector for all paths
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    polygons = list(
        converter.to_mpolygons(
            paths,
            distance=distance,
            segments=segments,
            extrusion=extrusion,
            dxfattribs=dxfattribs,
        )
    )
    for polygon in polygons:
        layout.add_entity(polygon)
    return EntityQuery(polygons)


def render_polylines3d(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    distance: float = MAX_DISTANCE,
    segments: int = MIN_SEGMENTS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as 3D
    :class:`~ezdxf.entities.Polyline` entities.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path`or :class:`Path2d` objects
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """

    polylines3d = list(
        converter.to_polylines3d(
            paths,
            distance=distance,
            segments=segments,
            dxfattribs=dxfattribs,
        )
    )
    for polyline3d in polylines3d:
        layout.add_entity(polyline3d)
    return EntityQuery(polylines3d)


def render_lines(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    distance: float = MAX_DISTANCE,
    segments: int = MIN_SEGMENTS,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as
    :class:`~ezdxf.entities.Line` entities.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path`or :class:`Path2d` objects
        distance:  maximum distance, see :meth:`Path.flattening`
        segments: minimum segment count per Bézier curve
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    lines = list(
        converter.to_lines(
            paths,
            distance=distance,
            segments=segments,
            dxfattribs=dxfattribs,
        )
    )
    for line in lines:
        layout.add_entity(line)
    return EntityQuery(lines)


def render_splines_and_polylines(
    layout: GenericLayoutType,
    paths: Iterable[Path],
    *,
    g1_tol: float = G1_TOL,
    dxfattribs=None,
) -> EntityQuery:
    """Render the given `paths` into `layout` as :class:`~ezdxf.entities.Spline`
    and 3D :class:`~ezdxf.entities.Polyline` entities.

    Args:
        layout: the modelspace, a paperspace layout or a block definition
        paths: iterable of :class:`Path`or :class:`Path2d` objects
        g1_tol: tolerance for G1 continuity check
        dxfattribs: additional DXF attribs

    Returns:
        created entities in an :class:`~ezdxf.query.EntityQuery` object

    """
    entities = list(
        converter.to_splines_and_polylines(
            paths,
            g1_tol=g1_tol,
            dxfattribs=dxfattribs,
        )
    )
    for entity in entities:
        layout.add_entity(entity)
    return EntityQuery(entities)


def add_ellipse(
    path: Path, ellipse: ConstructionEllipse, segments=1, reset=True
) -> None:
    """Add an elliptical arc as multiple cubic Bèzier-curves to the given
    `path`, use :meth:`~ezdxf.math.ConstructionEllipse.from_arc` constructor
    of class :class:`~ezdxf.math.ConstructionEllipse` to add circular arcs.

    Auto-detect the connection point to the given `path`, if neither the start-
    nor the end point of the ellipse is close to the path end point, a line from
    the path end point to the ellipse start point will be added automatically
    (see :func:`add_bezier4p`).

    By default, the start of an **empty** path is set to the start point of
    the ellipse, setting argument `reset` to ``False`` prevents this
    behavior.

    Args:
        path: :class:`~ezdxf.path.Path` object
        ellipse: ellipse parameters as :class:`~ezdxf.math.ConstructionEllipse`
            object
        segments: count of Bèzier-curve segments, at least one segment for
            each quarter (pi/2), ``1`` for as few as possible.
        reset: set start point to start of ellipse if path is empty

    """
    if abs(ellipse.param_span) < 1e-9:
        return
    if len(path) == 0 and reset:
        path.start = ellipse.start_point
    add_bezier4p(path, cubic_bezier_from_ellipse(ellipse, segments))


def add_bezier4p(path: Path, curves: Iterable[Bezier4P]) -> None:
    """Add multiple cubic Bèzier-curves to the given `path`.

    Auto-detect the connection point to the given `path`, if neither the start-
    nor the end point of the curves is close to the path end point, a line from
    the path end point to the start point of the first curve will be added
    automatically.

    """
    rel_tol = 1e-15
    abs_tol = 0.0
    curves = list(curves)
    if not len(curves):
        return
    end = curves[-1].control_points[-1]
    if path.end.isclose(end):
        # connect to new curves end point
        curves = reverse_bezier_curves(curves)

    for curve in curves:
        start, ctrl1, ctrl2, end = curve.control_points
        if not start.isclose(path.end):
            path.line_to(start)

        # add linear bezier segments as LINE_TO commands
        if start.isclose(ctrl1, rel_tol=rel_tol, abs_tol=abs_tol) and end.isclose(
            ctrl2, rel_tol=rel_tol, abs_tol=abs_tol
        ):
            path.line_to(end)
        else:
            path.curve4_to(end, ctrl1, ctrl2)


def add_bezier3p(path: Path, curves: Iterable[Bezier3P]) -> None:
    """Add multiple quadratic Bèzier-curves to the given `path`.

    Auto-detect the connection point to the given `path`, if neither the start-
    nor the end point of the curves is close to the path end point, a line from
    the path end point to the start point of the first curve will be added
    automatically.

    """
    rel_tol = 1e-15
    abs_tol = 0.0
    curves = list(curves)
    if not len(curves):
        return
    end = curves[-1].control_points[-1]
    if path.end.isclose(end):
        # connect to new curves end point
        curves = reverse_bezier_curves(curves)

    for curve in curves:
        start, ctrl, end = curve.control_points
        if not start.isclose(path.end, rel_tol=rel_tol, abs_tol=abs_tol):
            path.line_to(start)

        if start.isclose(ctrl, rel_tol=rel_tol, abs_tol=abs_tol) or end.isclose(
            ctrl, rel_tol=rel_tol, abs_tol=abs_tol
        ):
            path.line_to(end)
        else:
            path.curve3_to(end, ctrl)


def add_2d_polyline(
    path: Path,
    points: Iterable[Sequence[float]],
    close: bool,
    ocs: OCS,
    elevation: float,
    segments: int = 1,
) -> None:
    """Internal API to add 2D polylines which may include bulges to an
    **empty** path.

    """

    def bulge_to(p1: Vec3, p2: Vec3, bulge: float, segments: int):
        if p1.isclose(p2, rel_tol=IS_CLOSE_TOL, abs_tol=0):
            return
        # each cubic_bezier adds 3 segments, need 1 minimum
        num_bez = math.ceil(segments / 3)
        center, start_angle, end_angle, radius = bulge_to_arc(p1, p2, bulge)
        # normalize angles into range 0 .. 2pi
        start_angle = start_angle % math.tau
        end_angle = end_angle % math.tau
        if start_angle > end_angle:
            end_angle += math.tau
        angles = list(np.linspace(start_angle, end_angle, num_bez + 1))
        curves = []
        for i in range(num_bez):
            ellipse = ConstructionEllipse.from_arc(
                center,
                radius,
                Z_AXIS,
                math.degrees(angles[i]),
                math.degrees(angles[i + 1]),
            )
            curves.extend(list(cubic_bezier_from_ellipse(ellipse)))
        curve0 = curves[0]
        cp0 = curve0.control_points[0]
        if cp0.isclose(p2, rel_tol=IS_CLOSE_TOL, abs_tol=0):
            curves = reverse_bezier_curves(curves)
        add_bezier4p(path, curves)

    if len(path):
        raise ValueError("Requires an empty path.")

    prev_point: Optional[Vec3] = None
    prev_bulge: float = 0
    for x, y, bulge in points:
        # Bulge values near 0 but != 0 cause crashes! #329
        if abs(bulge) < 1e-6:
            bulge = 0
        point = Vec3(x, y)
        if prev_point is None:
            path.start = point
            prev_point = point
            prev_bulge = bulge
            continue

        if prev_bulge:
            bulge_to(prev_point, point, prev_bulge, segments)
        else:
            path.line_to(point)
        prev_point = point
        prev_bulge = bulge

    if close and not path.start.isclose(path.end, rel_tol=IS_CLOSE_TOL, abs_tol=0):
        if prev_bulge:
            bulge_to(path.end, path.start, prev_bulge, segments)
        else:
            path.line_to(path.start)

    if ocs.transform or elevation:
        path.to_wcs(ocs, elevation)


def add_spline(path: Path, spline: BSpline, level=4, reset=True) -> None:
    """Add a B-spline as multiple cubic Bèzier-curves.

    Non-rational B-splines of 3rd degree gets a perfect conversion to
    cubic Bézier curves with a minimal count of curve segments, all other
    B-spline require much more curve segments for approximation.

    Auto-detect the connection point to the given `path`, if neither the start-
    nor the end point of the B-spline is close to the path end point, a line
    from the path end point to the start point of the B-spline will be added
    automatically. (see :meth:`add_bezier4p`).

    By default, the start of an **empty** path is set to the start point of
    the spline, setting argument `reset` to ``False`` prevents this
    behavior.

    Args:
        path: :class:`~ezdxf.path.Path` object
        spline: B-spline parameters as :class:`~ezdxf.math.BSpline` object
        level: subdivision level of approximation segments
        reset: set start point to start of spline if path is empty

    """
    if len(path) == 0 and reset:
        path.start = spline.point(0)
    curves: Iterable[Bezier4P]
    if spline.degree == 3 and not spline.is_rational and spline.is_clamped:
        curves = [Bezier4P(points) for points in spline.bezier_decomposition()]
    else:
        curves = spline.cubic_bezier_approximation(level=level)
    add_bezier4p(path, curves)


def have_close_control_vertices(
    a: Path, b: Path, *, rel_tol=1e-9, abs_tol=1e-12
) -> bool:
    """Returns ``True`` if the control vertices of given paths are close."""
    return all(
        cp_a.isclose(cp_b, rel_tol=rel_tol, abs_tol=abs_tol)
        for cp_a, cp_b in zip(a.control_vertices(), b.control_vertices())
    )


def lines_to_curve3(path: Path) -> Path:
    """Replaces all lines by quadratic Bézier curves.
    Returns a new :class:`Path` instance.
    """
    return _all_lines_to_curve(path, count=3)


def lines_to_curve4(path: Path) -> Path:
    """Replaces all lines by cubic Bézier curves.
    Returns a new :class:`Path` instance.
    """
    return _all_lines_to_curve(path, count=4)


def _all_lines_to_curve(path: Path, count: int = 4) -> Path:
    assert count == 4 or count == 3, f"invalid count: {count}"

    cmds = path.commands()
    size = len(cmds)
    if size == 0:  # empty path
        return Path()
    start = path.start
    line_to = Command.LINE_TO
    new_path = Path(path.start)
    for cmd in cmds:
        if cmd.type == line_to:
            if start.isclose(cmd.end):
                if size == 1:
                    # Path has only one LINE_TO command which should not be
                    # removed:
                    # 1. may represent a point
                    # 2. removing the last segment turns the path into
                    #    an empty path - unexpected behavior?
                    new_path.append_path_element(cmd)
                    return new_path
                # else remove line segment (start==end)
            else:
                vertices = linear_vertex_spacing(start, cmd.end, count)
                if count == 3:
                    new_path.curve3_to(vertices[2], ctrl=vertices[1])
                else:  # count == 4
                    new_path.curve4_to(
                        vertices[3],
                        ctrl1=vertices[1],
                        ctrl2=vertices[2],
                    )
        else:
            new_path.append_path_element(cmd)
        start = cmd.end
    return new_path


def _get_local_fillet_ucs(p0, p1, p2, radius) -> tuple[Vec3, float, UCS]:
    dir1 = (p0 - p1).normalize()
    dir2 = (p2 - p1).normalize()
    if dir1.isclose(dir2) or dir1.isclose(-dir2):
        raise ZeroDivisionError

    # arc start- and end points:
    angle = dir1.angle_between(dir2)
    tangent_length = inscribe_circle_tangent_length(dir1, dir2, radius)
    # starting point of the fillet arc
    arc_start_point = p1 + (dir1 * tangent_length)

    # create local coordinate system:
    # origin = center of the fillet arc
    # x-axis = arc_center -> arc_start_point
    local_z_axis = dir2.cross(dir1)
    # radius_vec points from arc_start_point to the center of the fillet arc
    radius_vec = local_z_axis.cross(-dir1).normalize(radius)
    arc_center = arc_start_point + radius_vec
    ucs = UCS(origin=arc_center, ux=-radius_vec, uz=local_z_axis)
    return arc_start_point, math.pi - angle, ucs


def fillet(points: Sequence[Vec3], radius: float) -> Path:
    """Returns a :class:`Path` with circular fillets of given `radius` between
    straight line segments.

    Args:
        points: coordinates of the line segments
        radius: fillet radius

    """
    if len(points) < 3:
        raise ValueError("at least 3 not coincident points required")
    if radius <= 0:
        raise ValueError(f"invalid radius: {radius}")
    lines = [(p0, p1) for p0, p1 in zip(points, points[1:])]
    p = Path(points[0])
    for (p0, p1), (p2, p3) in zip(lines, lines[1:]):
        try:
            start_point, angle, ucs = _get_local_fillet_ucs(p0, p1, p3, radius)
        except ZeroDivisionError:
            p.line_to(p1)
            continue

        # add path elements:
        p.line_to(start_point)
        for params in cubic_bezier_arc_parameters(0, angle):
            # scale arc parameters by radius:
            bez_points = tuple(ucs.points_to_wcs(v * radius for v in params))
            p.curve4_to(bez_points[-1], bez_points[1], bez_points[2])
    p.line_to(points[-1])
    return p


def _segment_count(angle: float, count: int) -> int:
    count = max(4, count)
    return max(int(angle / (math.tau / count)), 1)


def polygonal_fillet(points: Sequence[Vec3], radius: float, count: int = 32) -> Path:
    """
    Returns a :class:`Path` with polygonal fillets of given `radius` between
    straight line segments. The `count` argument defines the vertex count of the
    fillet for a full circle.

    Args:
        points: coordinates of the line segments
        radius: fillet radius
        count: polygon vertex count for a full circle, minimum is 4

    """
    if len(points) < 3:
        raise ValueError("at least 3 not coincident points required")
    if radius <= 0:
        raise ValueError(f"invalid radius: {radius}")
    lines = [(p0, p1) for p0, p1 in zip(points, points[1:])]
    p = Path(points[0])
    for (p0, p1), (p2, p3) in zip(lines, lines[1:]):
        try:
            _, angle, ucs = _get_local_fillet_ucs(p0, p1, p3, radius)
        except ZeroDivisionError:
            p.line_to(p1)
            continue
        segments = _segment_count(angle, count)
        delta = angle / segments
        # add path elements:
        for i in range(segments + 1):
            radius_vec = Vec3.from_angle(i * delta, radius)
            p.line_to(ucs.to_wcs(radius_vec))

    p.line_to(points[-1])
    return p


def chamfer(points: Sequence[Vec3], length: float) -> Path:
    """
    Returns a :class:`Path` with chamfers of given `length` between
    straight line segments.

    Args:
        points: coordinates of the line segments
        length: chamfer length

    """
    if len(points) < 3:
        raise ValueError("at least 3 not coincident points required")
    lines = [(p0, p1) for p0, p1 in zip(points, points[1:])]
    p = Path(points[0])
    for (p0, p1), (p2, p3) in zip(lines, lines[1:]):
        # p1 is p2 !
        try:
            dir1 = (p0 - p1).normalize()
            dir2 = (p3 - p2).normalize()
            if dir1.isclose(dir2) or dir1.isclose(-dir2):
                raise ZeroDivisionError
            angle = dir1.angle_between(dir2) / 2.0
            a = abs((length / 2.0) / math.sin(angle))
        except ZeroDivisionError:
            p.line_to(p1)
            continue
        p.line_to(p1 + (dir1 * a))
        p.line_to(p2 + (dir2 * a))
    p.line_to(points[-1])
    return p


def chamfer2(points: Sequence[Vec3], a: float, b: float) -> Path:
    """
    Returns a :class:`Path` with chamfers at the given distances `a` and `b`
    from the segment points between straight line segments.

    Args:
        points: coordinates of the line segments
        a: distance of the chamfer start point to the segment point
        b: distance of the chamfer end point to the segment point

    """
    if len(points) < 3:
        raise ValueError("at least 3 non-coincident points required")
    lines = [(p0, p1) for p0, p1 in zip(points, points[1:])]
    p = Path(points[0])
    for (p0, p1), (p2, p3) in zip(lines, lines[1:]):
        # p1 is p2 !
        try:
            dir1 = (p0 - p1).normalize()
            dir2 = (p3 - p2).normalize()
            if dir1.isclose(dir2) or dir1.isclose(-dir2):
                raise ZeroDivisionError
        except ZeroDivisionError:
            p.line_to(p1)
            continue
        p.line_to(p1 + (dir1 * a))
        p.line_to(p2 + (dir2 * b))
    p.line_to(points[-1])
    return p


def triangulate(
    paths: Iterable[Path], max_sagitta: float = 0.01, min_segments: int = 16
) -> Iterator[Sequence[Vec2]]:
    """Tessellate nested 2D paths into triangle-faces. For 3D paths the
    projection onto the xy-plane will be triangulated.

    Args:
        paths: iterable of nested Path instances
        max_sagitta: maximum distance from the center of the curve to the
            center of the line segment between two approximation points to determine if
            a segment should be subdivided.
        min_segments: minimum segment count per Bézier curve

    """
    for polygon in nesting.group_paths(single_paths(paths)):
        exterior = polygon[0].flattening(max_sagitta, min_segments)
        holes = [p.flattening(max_sagitta, min_segments) for p in polygon[1:]]
        yield from mapbox_earcut_2d(exterior, holes)


def is_rectangular(path: Path, aligned=True) -> bool:
    """Returns ``True`` if `path` is a rectangular quadrilateral (square or
    rectangle). If the argument `aligned` is ``True`` all sides of the
    quadrilateral have to be parallel to the x- and y-axis.
    """
    points = path.control_vertices()
    if len(points) < 4:
        return False
    if points[0].isclose(points[-1]):
        points.pop()
    if len(points) != 4:
        return False

    if aligned:
        first_side = points[1] - points[0]
        if not (abs(first_side.x) < 1e-12 or abs(first_side.y) < 1e-12):
            return False

    # horizontal sides
    v1 = points[0].distance(points[1])
    v2 = points[2].distance(points[3])
    if not math.isclose(v1, v2):
        return False
    # vertical sides
    v1 = points[1].distance(points[2])
    v2 = points[3].distance(points[0])
    if not math.isclose(v1, v2):
        return False
    # diagonals
    v1 = points[0].distance(points[2])
    v2 = points[1].distance(points[3])
    if not math.isclose(v1, v2):
        return False

    return True



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
import dataclasses
import enum
import copy
from ezdxf.colors import RGB


class FillType(enum.IntEnum):
    """Fill type enumeration."""

    NONE = 0
    SOLID = 1
    HATCHING = 2
    CROSS_HATCHING = 3
    SHADING = 4


class FillMethod(enum.IntEnum):
    """Fill method enumeration."""

    EVEN_ODD = 0
    NON_ZERO_WINDING = 1


RGB_NONE = RGB(-1, -1, -1)
RGB_BLACK = RGB(0, 0, 0)
RGB_WHITE = RGB(255, 255, 255)
LIGHT_GREY = RGB(200, 200, 200)


@dataclasses.dataclass
class Pen:
    """Represents a pen table entry."""

    index: int
    width: float  # in mm
    color: RGB


class Properties:
    """Consolidated display properties."""

    DEFAULT_PEN = Pen(1, 0.35, RGB_NONE)

    def __init__(self) -> None:
        # hashed content
        self.pen_index: int = 1
        self.pen_color = RGB_NONE
        self.pen_width: float = 0.35  # in mm
        self.fill_type = FillType.SOLID
        self.fill_method = FillMethod.EVEN_ODD
        self.fill_hatch_line_angle: float = 0.0  # in degrees
        self.fill_hatch_line_spacing: float = 40.0  # in plotter units
        self.fill_shading_density: float = 100.0
        # not hashed content
        self.max_pen_count: int = 2
        self.pen_table: dict[int, Pen] = {}
        self.reset()

    def hash(self) -> int:
        return hash(
            (
                self.pen_index,
                self.pen_color,
                self.pen_width,
                self.fill_type,
                self.fill_method,
                self.fill_hatch_line_angle,
                self.fill_hatch_line_spacing,
                self.fill_shading_density,
            )
        )

    def copy(self) -> Properties:
        # the pen table is shared across all copies of Properties
        return copy.copy(self)

    def setup_default_pen_table(self):
        if len(self.pen_table):
            return
        pens = self.pen_table
        width = self.DEFAULT_PEN.width
        pens[0] = Pen(0, width, RGB(255, 255, 255))  # white
        pens[1] = Pen(1, width, RGB(0, 0, 0))  # black
        pens[2] = Pen(2, width, RGB(255, 0, 0))  # red
        pens[3] = Pen(3, width, RGB(0, 255, 0))  # green
        pens[4] = Pen(4, width, RGB(255, 255, 0))  # yellow
        pens[5] = Pen(5, width, RGB(0, 0, 255))  # blue
        pens[6] = Pen(6, width, RGB(255, 0, 255))  # magenta
        pens[7] = Pen(6, width, RGB(0, 255, 255))  # cyan

    def reset(self) -> None:
        self.max_pen_count = 2
        self.pen_index = self.DEFAULT_PEN.index
        self.pen_color = self.DEFAULT_PEN.color
        self.pen_width = self.DEFAULT_PEN.width
        self.pen_table = {}
        self.fill_type = FillType.SOLID
        self.fill_method = FillMethod.EVEN_ODD
        self.fill_hatch_line_angle = 0.0
        self.fill_hatch_line_spacing = 40.0
        self.fill_shading_density = 1.0
        self.setup_default_pen_table()

    def get_pen(self, index: int) -> Pen:
        return self.pen_table.get(index, self.DEFAULT_PEN)

    def set_max_pen_count(self, count: int) -> None:
        self.max_pen_count = count

    def set_current_pen(self, index: int) -> None:
        self.pen_index = index
        pen = self.get_pen(index)
        self.pen_width = pen.width
        self.pen_color = pen.color

    def set_pen_width(self, index: int, width: float) -> None:
        if index == -1:
            self.pen_width = width
        else:
            pen = self.pen_table.setdefault(
                index, Pen(index, width, self.DEFAULT_PEN.color)
            )
            pen.width = width

    def set_pen_color(self, index: int, rgb: RGB) -> None:
        if index == -1:
            self.pen_color = rgb
        else:
            pen = self.pen_table.setdefault(
                index, Pen(index, self.DEFAULT_PEN.width, rgb)
            )
            pen.color = rgb

    def set_fill_type(self, fill_type: int, spacing: float, angle: float):
        if fill_type == 3:
            self.fill_type = FillType.HATCHING
            self.fill_hatch_line_spacing = spacing
            self.fill_hatch_line_angle = angle
        elif fill_type == 4:
            self.fill_type = FillType.CROSS_HATCHING
            self.fill_hatch_line_spacing = spacing
            self.fill_hatch_line_angle = angle
        elif fill_type == 10:
            self.fill_type = FillType.SHADING
            self.fill_shading_density = spacing
        else:
            self.fill_type = FillType.SOLID

    def set_fill_method(self, fill_method: int) -> None:
        self.fill_method = FillMethod(fill_method)

    def resolve_pen_color(self) -> RGB:
        """Returns the final RGB pen color."""
        rgb = self.pen_color
        if rgb is RGB_NONE:
            pen = self.pen_table.get(self.pen_index, self.DEFAULT_PEN)
            rgb = pen.color
        if rgb is RGB_NONE:
            return RGB_BLACK
        return rgb

    def resolve_fill_color(self) -> RGB:
        """Returns the final RGB fill color."""
        ft = self.fill_type
        if ft == FillType.SOLID:
            return self.resolve_pen_color()
        elif ft == FillType.SHADING:
            grey = min(int(2.55 * (100.0 - self.fill_shading_density)), 255)
            return RGB(grey, grey, grey)
        elif ft == FillType.HATCHING or ft == FillType.CROSS_HATCHING:
            return LIGHT_GREY
        return RGB_WHITE



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import Vec3
from ezdxf.math.clustering import k_means, average_cluster_radius, average_intra_cluster_distance

POINTS = [Vec3.random(100) for _ in range(100)]


@pytest.mark.parametrize("k", [4, 5, 6])
def test_cluster_random_points(k):
    clusters = k_means(POINTS, k, max_iter=5)
    assert len(clusters) == k
    assert sum(map(len, clusters)) == 100


def test_measure_average_cluster_radius():
    clusters = k_means(POINTS, 5, max_iter=5)
    assert average_cluster_radius(clusters) > 10.0


def test_measure_average_intra_cluster_distance():
    clusters = k_means(POINTS, 5, max_iter=5)
    assert average_intra_cluster_distance(clusters) > 10.0


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2016-2023, Manfred Moitzi
# License: MIT License
import re
import codecs
import binascii

surrogate_escape = codecs.lookup_error("surrogateescape")
BACKSLASH_UNICODE = re.compile(r"(\\U\+[A-F0-9]{4})")
MIF_ENCODED = re.compile(r"(\\M\+[1-5][A-F0-9]{4})")


def dxf_backslash_replace(exc: Exception):
    if isinstance(exc, (UnicodeEncodeError, UnicodeTranslateError)):
        s = ""
        # mypy does not recognize properties: exc.start, exc.end, exc.object
        for c in exc.object[exc.start : exc.end]:
            x = ord(c)
            if x <= 0xFF:
                s += "\\x%02x" % x
            elif 0xDC80 <= x <= 0xDCFF:
                # Delegate surrogate handling:
                return surrogate_escape(exc)
            elif x <= 0xFFFF:
                s += "\\U+%04x" % x
            else:
                s += "\\U+%08x" % x
        return s, exc.end
    else:
        raise TypeError(f"Can't handle {exc.__class__.__name__}")


def encode(s: str, encoding="utf8") -> bytes:
    """Shortcut to use the correct error handler"""
    return s.encode(encoding, errors="dxfreplace")


def _decode(s: str) -> str:
    if s.startswith(r"\U+"):
        return chr(int(s[3:], 16))
    else:
        return s


def has_dxf_unicode(s: str) -> bool:
    """Returns ``True``  if string `s` contains ``\\U+xxxx`` encoded characters."""
    return bool(re.search(BACKSLASH_UNICODE, s))


def decode_dxf_unicode(s: str) -> str:
    """Decode ``\\U+xxxx`` encoded characters."""

    return "".join(_decode(part) for part in re.split(BACKSLASH_UNICODE, s))


def has_mif_encoding(s: str) -> bool:
    """Returns ``True`` if string `s` contains MIF encoded (``\\M+cxxxx``) characters.
    """
    return bool(re.search(MIF_ENCODED, s))


def decode_mif_to_unicode(s: str) -> str:
    """Decode MIF encoded characters ``\\M+cxxxx``."""
    return "".join(_decode_mif(part) for part in re.split(MIF_ENCODED, s))


MIF_CODE_PAGE = {
    # See https://docs.intellicad.org/files/oda/2021_11/oda_drawings_docs/frames.html?frmname=topic&frmfile=FontHandling.html
    "1": "cp932",  # Japanese (Shift-JIS)
    "2": "cp950",  # Traditional Chinese (Big 5)
    "3": "cp949",  # Wansung (KS C-5601-1987)
    "4": "cp1391",  # Johab (KS C-5601-1992)
    "5": "cp936",  # Simplified Chinese (GB 2312-80)
}


def _decode_mif(s: str) -> str:
    if s.startswith(r"\M+"):
        try:
            code_page = MIF_CODE_PAGE[s[3]]
            codec = codecs.lookup(code_page)
            byte_data = binascii.unhexlify(s[4:])
            return codec.decode(byte_data)[0]
        except Exception:
            pass
    return s



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Any, TYPE_CHECKING

import math
import os
import pathlib

from ezdxf.addons import xplayer
from ezdxf.addons.xqt import QtWidgets, QtGui, QtCore, QMessageBox
from ezdxf.addons.drawing import svg, layout, pymupdf, dxf
from ezdxf.addons.drawing.qtviewer import CADGraphicsView
from ezdxf.addons.drawing.pyqt import PyQtPlaybackBackend

from . import api
from .deps import BoundingBox2d, Matrix44, colors

if TYPE_CHECKING:
    from ezdxf.document import Drawing

VIEWER_NAME = "HPGL/2 Viewer"


class HPGL2Widget(QtWidgets.QWidget):
    def __init__(self, view: CADGraphicsView) -> None:
        super().__init__()
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(view)
        self.setLayout(layout)
        self._view = view
        self._view.closing.connect(self.close)
        self._player: api.Player = api.Player([], {})
        self._reset_backend()

    def _reset_backend(self) -> None:
        self._backend = PyQtPlaybackBackend()

    @property
    def view(self) -> CADGraphicsView:
        return self._view

    @property
    def player(self) -> api.Player:
        return self._player.copy()

    def plot(self, data: bytes) -> None:
        self._reset_backend()
        self._player = api.record_plotter_output(data, api.MergeControl.AUTO)

    def replay(
        self, bg_color="#ffffff", override=None, reset_view: bool = True
    ) -> None:
        self._reset_backend()
        self._view.begin_loading()
        new_scene = QtWidgets.QGraphicsScene()
        self._backend.set_scene(new_scene)
        new_scene.addItem(self._bg_paper(bg_color))

        xplayer.hpgl2_to_drawing(
            self._player, self._backend, bg_color="", override=override
        )
        self._view.end_loading(new_scene)
        self._view.buffer_scene_rect()
        if reset_view:
            self._view.fit_to_scene()

    def _bg_paper(self, color):
        bbox = self._player.bbox()
        insert = bbox.extmin
        size = bbox.size
        rect = QtWidgets.QGraphicsRectItem(insert.x, insert.y, size.x, size.y)
        rect.setBrush(QtGui.QBrush(QtGui.QColor(color)))
        return rect


SPACING = 20
DEFAULT_DPI = 96
COLOR_SCHEMA = [
    "Default",
    "Black on White",
    "White on Black",
    "Monochrome Light",
    "Monochrome Dark",
    "Blueprint High Contrast",
    "Blueprint Low Contrast",
]


class HPGL2Viewer(QtWidgets.QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self._cad = HPGL2Widget(CADGraphicsView())
        self._view = self._cad.view
        self._player: api.Player = api.Player([], {})
        self._bbox: BoundingBox2d = BoundingBox2d()
        self._page_rotation = 0
        self._color_scheme = 0
        self._current_file = pathlib.Path()

        self.page_size_label = QtWidgets.QLabel()
        self.png_size_label = QtWidgets.QLabel()
        self.message_label = QtWidgets.QLabel()
        self.scaling_factor_line_edit = QtWidgets.QLineEdit("1")
        self.dpi_line_edit = QtWidgets.QLineEdit(str(DEFAULT_DPI))

        self.flip_x_check_box = QtWidgets.QCheckBox("Horizontal")
        self.flip_x_check_box.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.flip_x_check_box.stateChanged.connect(self.update_view)

        self.flip_y_check_box = QtWidgets.QCheckBox("Vertical")
        self.flip_y_check_box.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.flip_y_check_box.stateChanged.connect(self.update_view)

        self.rotation_combo_box = QtWidgets.QComboBox()
        self.rotation_combo_box.addItems(["0", "90", "180", "270"])
        self.rotation_combo_box.currentIndexChanged.connect(self.update_rotation)

        self.color_combo_box = QtWidgets.QComboBox()
        self.color_combo_box.addItems(COLOR_SCHEMA)
        self.color_combo_box.currentIndexChanged.connect(self.update_colors)

        self.aci_export_mode = QtWidgets.QCheckBox("ACI Export Mode")
        self.aci_export_mode.setCheckState(QtCore.Qt.CheckState.Unchecked)

        self.export_svg_button = QtWidgets.QPushButton("Export SVG")
        self.export_svg_button.clicked.connect(self.export_svg)
        self.export_png_button = QtWidgets.QPushButton("Export PNG")
        self.export_png_button.clicked.connect(self.export_png)
        self.export_pdf_button = QtWidgets.QPushButton("Export PDF")
        self.export_pdf_button.clicked.connect(self.export_pdf)
        self.export_dxf_button = QtWidgets.QPushButton("Export DXF")
        self.export_dxf_button.clicked.connect(self.export_dxf)
        self.disable_export_buttons(True)

        self.scaling_factor_line_edit.editingFinished.connect(self.update_sidebar)
        self.dpi_line_edit.editingFinished.connect(self.update_sidebar)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        container = QtWidgets.QWidget()
        container.setLayout(layout)
        self.setCentralWidget(container)

        layout.addWidget(self._cad)
        sidebar = self.make_sidebar()
        layout.addWidget(sidebar)
        self.setWindowTitle(VIEWER_NAME)
        self.resize(1600, 900)
        self.show()

    def reset_values(self):
        self.scaling_factor_line_edit.setText("1")
        self.dpi_line_edit.setText(str(DEFAULT_DPI))
        self.flip_x_check_box.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.flip_y_check_box.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.rotation_combo_box.setCurrentIndex(0)
        self._page_rotation = 0
        self.update_view()

    def make_sidebar(self) -> QtWidgets.QWidget:
        sidebar = QtWidgets.QWidget()
        v_layout = QtWidgets.QVBoxLayout()
        v_layout.setContentsMargins(SPACING // 2, 0, SPACING // 2, 0)
        sidebar.setLayout(v_layout)

        policy = QtWidgets.QSizePolicy()
        policy.setHorizontalPolicy(QtWidgets.QSizePolicy.Policy.Fixed)
        sidebar.setSizePolicy(policy)

        open_button = QtWidgets.QPushButton("Open HPGL/2 File")
        open_button.clicked.connect(self.select_plot_file)
        v_layout.addWidget(open_button)
        v_layout.addWidget(self.page_size_label)
        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(QtWidgets.QLabel("Scaling Factor:"))
        h_layout.addWidget(self.scaling_factor_line_edit)
        v_layout.addLayout(h_layout)

        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(QtWidgets.QLabel("Page Rotation:"))
        h_layout.addWidget(self.rotation_combo_box)
        v_layout.addLayout(h_layout)

        group = QtWidgets.QGroupBox("Mirror Page")
        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(self.flip_x_check_box)
        h_layout.addWidget(self.flip_y_check_box)
        group.setLayout(h_layout)
        v_layout.addWidget(group)

        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(QtWidgets.QLabel("Colors:"))
        h_layout.addWidget(self.color_combo_box)
        v_layout.addLayout(h_layout)

        v_layout.addSpacing(SPACING)

        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(QtWidgets.QLabel("DPI (PNG only):"))
        h_layout.addWidget(self.dpi_line_edit)
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.png_size_label)

        v_layout.addWidget(self.export_png_button)
        v_layout.addWidget(self.export_svg_button)
        v_layout.addWidget(self.export_pdf_button)

        v_layout.addSpacing(SPACING)

        v_layout.addWidget(self.aci_export_mode)
        v_layout.addWidget(self.export_dxf_button)

        v_layout.addSpacing(SPACING)

        reset_button = QtWidgets.QPushButton("Reset")
        reset_button.clicked.connect(self.reset_values)
        v_layout.addWidget(reset_button)

        v_layout.addSpacing(SPACING)

        v_layout.addWidget(self.message_label)
        return sidebar

    def disable_export_buttons(self, disabled: bool):
        self.export_svg_button.setDisabled(disabled)
        self.export_dxf_button.setDisabled(disabled)
        if pymupdf.is_pymupdf_installed:
            self.export_png_button.setDisabled(disabled)
            self.export_pdf_button.setDisabled(disabled)
        else:
            print("PDF/PNG export requires the PyMuPdf package!")
            self.export_png_button.setDisabled(True)
            self.export_pdf_button.setDisabled(True)

    def load_plot_file(self, path: str | os.PathLike, force=False) -> None:
        try:
            with open(path, "rb") as fp:
                data = fp.read()
            if force:
                data = b"%1B" + data
            self.set_plot_data(data, path)
        except IOError as e:
            QtWidgets.QMessageBox.critical(self, "Loading Error", str(e))

    def select_plot_file(self) -> None:
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            dir=str(self._current_file.parent),
            caption="Select HPGL/2 Plot File",
            filter="Plot Files (*.plt)",
        )
        if path:
            self.load_plot_file(path)

    def set_plot_data(self, data: bytes, filename: str | os.PathLike) -> None:
        try:
            self._cad.plot(data)
        except api.Hpgl2Error as e:
            msg = f"Cannot plot HPGL/2 file '{filename}', {str(e)}"
            QtWidgets.QMessageBox.critical(self, "Plot Error", msg)
            return
        self._player = self._cad.player
        self._bbox = self._player.bbox()
        self._current_file = pathlib.Path(filename)
        self.update_colors(self._color_scheme)
        self.update_sidebar()
        self.setWindowTitle(f"{VIEWER_NAME} - " + str(filename))
        self.disable_export_buttons(False)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        self._view.fit_to_scene()

    def get_scale_factor(self) -> float:
        try:
            return float(self.scaling_factor_line_edit.text())
        except ValueError:
            return 1.0

    def get_dpi(self) -> int:
        try:
            return int(self.dpi_line_edit.text())
        except ValueError:
            return DEFAULT_DPI

    def get_page_size(self) -> tuple[int, int]:
        factor = self.get_scale_factor()
        x = 0
        y = 0
        if self._bbox.has_data:
            size = self._bbox.size
            # 40 plot units = 1mm
            x = round(size.x / 40 * factor)
            y = round(size.y / 40 * factor)
        if self._page_rotation in (90, 270):
            x, y = y, x
        return x, y

    def get_pixel_size(self) -> tuple[int, int]:
        dpi = self.get_dpi()
        x, y = self.get_page_size()
        return round(x / 25.4 * dpi), round(y / 25.4 * dpi)

    def get_flip_x(self) -> bool:
        return self.flip_x_check_box.checkState() == QtCore.Qt.CheckState.Checked

    def get_flip_y(self) -> bool:
        return self.flip_y_check_box.checkState() == QtCore.Qt.CheckState.Checked

    def get_aci_export_mode(self) -> bool:
        return self.aci_export_mode.checkState() == QtCore.Qt.CheckState.Checked

    def update_sidebar(self):
        x, y = self.get_page_size()
        self.page_size_label.setText(f"Page Size: {x}x{y}mm")
        px, py = self.get_pixel_size()
        self.png_size_label.setText(f"PNG Size: {px}x{py}px")
        self.clear_message()

    def update_view(self):
        self._view.setTransform(self.view_transformation())
        self._view.fit_to_scene()
        self.update_sidebar()

    def update_rotation(self, index: int):
        rotation = index * 90
        if rotation != self._page_rotation:
            self._page_rotation = rotation
            self.update_view()

    def update_colors(self, index: int):
        self._color_scheme = index
        self._cad.replay(*replay_properties(index))
        self.update_view()

    def view_transformation(self):
        if self._page_rotation == 0:
            m = Matrix44()
        else:
            m = Matrix44.z_rotate(math.radians(self._page_rotation))
        sx = -1 if self.get_flip_x() else 1
        # inverted y-axis
        sy = 1 if self.get_flip_y() else -1
        m @= Matrix44.scale(sx, sy, 1)
        return QtGui.QTransform(*m.get_2d_transformation())

    def show_message(self, msg: str) -> None:
        self.message_label.setText(msg)

    def clear_message(self) -> None:
        self.message_label.setText("")

    def get_export_name(self, suffix: str) -> str:
        return str(self._current_file.with_suffix(suffix))

    def export_svg(self) -> None:
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            dir=self.get_export_name(".svg"),
            caption="Save SVG File",
            filter="SVG Files (*.svg)",
        )
        if not path:
            return
        try:
            with open(path, "wt") as fp:
                fp.write(self.make_svg_string())
            self.show_message("SVG successfully exported")
        except IOError as e:
            QMessageBox.critical(self, "Export Error", str(e))

    def get_export_matrix(self) -> Matrix44:
        scale = self.get_scale_factor()
        rotation = self._page_rotation
        sx = -scale if self.get_flip_x() else scale
        sy = -scale if self.get_flip_y() else scale
        if rotation in (90, 270):
            sx, sy = sy, sx
        m = Matrix44.scale(sx, sy, 1)
        if rotation:
            m @= Matrix44.z_rotate(math.radians(rotation))
        return m

    def make_svg_string(self) -> str:
        """Replays the HPGL/2 recordings on the SVGBackend of the drawing add-on."""
        player = self._player.copy()
        player.transform(self.get_export_matrix())
        size = player.bbox().size
        svg_backend = svg.SVGBackend()
        bg_color, override = replay_properties(self._color_scheme)
        xplayer.hpgl2_to_drawing(
            player, svg_backend, bg_color=bg_color, override=override
        )
        del player  # free memory as soon as possible
        # 40 plot units == 1mm
        page = layout.Page(width=size.x / 40, height=size.y / 40)
        return svg_backend.get_string(page)

    def export_pdf(self) -> None:
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            dir=self.get_export_name(".pdf"),
            caption="Save PDF File",
            filter="PDF Files (*.pdf)",
        )
        if not path:
            return
        try:
            with open(path, "wb") as fp:
                fp.write(self._pymupdf_export(fmt="pdf"))
            self.show_message("PDF successfully exported")
        except IOError as e:
            QMessageBox.critical(self, "Export Error", str(e))

    def export_png(self) -> None:
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            dir=self.get_export_name(".png"),
            caption="Save PNG File",
            filter="PNG Files (*.png)",
        )
        if not path:
            return
        try:
            with open(path, "wb") as fp:
                fp.write(self._pymupdf_export(fmt="png"))
            self.show_message("PNG successfully exported")
        except IOError as e:
            QMessageBox.critical(self, "Export Error", str(e))

    def _pymupdf_export(self, fmt: str) -> bytes:
        """Replays the HPGL/2 recordings on the PyMuPdfBackend of the drawing add-on."""
        player = self._player.copy()
        player.transform(self.get_export_matrix())
        size = player.bbox().size
        pdf_backend = pymupdf.PyMuPdfBackend()
        bg_color, override = replay_properties(self._color_scheme)
        xplayer.hpgl2_to_drawing(
            player, pdf_backend, bg_color=bg_color, override=override
        )
        del player  # free memory as soon as possible
        # 40 plot units == 1mm
        page = layout.Page(width=size.x / 40, height=size.y / 40)
        if fmt == "pdf":
            return pdf_backend.get_pdf_bytes(page)
        else:
            return pdf_backend.get_pixmap_bytes(page, fmt=fmt, dpi=self.get_dpi())

    def export_dxf(self) -> None:
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            dir=self.get_export_name(".dxf"),
            caption="Save DXF File",
            filter="DXF Files (*.dxf)",
        )
        if not path:
            return
        doc = self._get_dxf_document()
        try:
            doc.saveas(path)
            self.show_message("DXF successfully exported")
        except IOError as e:
            QMessageBox.critical(self, "Export Error", str(e))

    def _get_dxf_document(self) -> Drawing:
        import ezdxf
        from ezdxf import zoom

        color_mode = (
            dxf.ColorMode.ACI if self.get_aci_export_mode() else dxf.ColorMode.RGB
        )

        doc = ezdxf.new()
        msp = doc.modelspace()
        player = self._player.copy()
        bbox = player.bbox()

        m = self.get_export_matrix()
        corners = m.fast_2d_transform(bbox.rect_vertices())
        # move content to origin:
        tx, ty = BoundingBox2d(corners).extmin
        m @= Matrix44.translate(-tx, -ty, 0)
        player.transform(m)
        bbox = player.bbox()

        dxf_backend = dxf.DXFBackend(msp, color_mode=color_mode)
        bg_color, override = replay_properties(self._color_scheme)
        if color_mode == dxf.ColorMode.RGB:
            doc.layers.add("BACKGROUND")
            bg = dxf.add_background(msp, bbox, colors.RGB.from_hex(bg_color))
            bg.dxf.layer = "BACKGROUND"
        # exports the HPGL/2 content in plot units (plu) as modelspace:
        # 1 plu = 0.025mm or 40 plu == 1mm
        xplayer.hpgl2_to_drawing(
            player, dxf_backend, bg_color=bg_color, override=override
        )
        del player
        if bbox.has_data:  # non-empty page
            zoom.window(msp, bbox.extmin, bbox.extmax)
            dxf.update_extents(doc, bbox)
            # paperspace is set up in mm:
            dxf.setup_paperspace(doc, bbox)
        return doc


def replay_properties(index: int) -> tuple[str, Any]:
    bg_color, override = "#ffffff", None  # default
    if index == 1:  # black on white
        bg_color, override = "#ffffff", xplayer.map_color("#000000")
    elif index == 2:  # white on black
        bg_color, override = "#000000", xplayer.map_color("#ffffff")
    elif index == 3:  # monochrome light
        bg_color, override = "#ffffff", xplayer.map_monochrome(dark_mode=False)
    elif index == 4:  # monochrome dark
        bg_color, override = "#000000", xplayer.map_monochrome(dark_mode=True)
    elif index == 5:  # blueprint high contrast
        bg_color, override = "#192c64", xplayer.map_color("#e9ebf3")
    elif index == 6:  # blueprint low contrast
        bg_color, override = "#243f8f", xplayer.map_color("#bdc5dd")
    return bg_color, override



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from pathlib import Path
import pytest
import ezdxf

DATA = Path(__file__).parent / "data"
GROUPS = "groups.dxf"


@pytest.fixture
def doc():
    return ezdxf.readfile(DATA / GROUPS)


def test_load_dxf_with_groups(doc):
    assert len(doc.groups) == 2
    (name1, g1), (name2, g2) = doc.groups
    assert name1 == "G1"
    assert len(g1) == 5
    assert name2 == "G2"
    assert len(g2) == 8


def delete_G1_content(dxf):
    g1 = dxf.groups.get("G1")
    for entity in g1:
        entity.destroy()


def test_delete_group_content_and_export_dxf(doc, tmp_path):
    delete_G1_content(doc)
    doc.saveas(tmp_path / GROUPS)
    assert True is True, "saveas() should not raise an exception"


def test_reload_dxf_with_empty_group_content(doc, tmp_path):
    delete_G1_content(doc)
    doc.saveas(tmp_path / GROUPS)
    del doc

    reload = ezdxf.readfile(tmp_path / GROUPS)
    assert len(reload.groups) == 2
    (name1, g1), (name2, g2) = reload.groups
    assert name1 == "G1"
    assert len(g1) == 0, "empty group still exist"
    assert name2 == "G2"
    assert len(g2) == 8


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf import appsettings


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_set_current_layer(doc):
    appsettings.set_current_layer(doc, "0")
    assert doc.header["$CLAYER"] == "0"


def test_invalid_layer_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_layer(doc, "INVALID")


def test_set_current_aci_color(doc):
    appsettings.set_current_color(doc, 7)
    assert doc.header["$CECOLOR"] == 7


def test_invalid_aci_color_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_color(doc, 300)


def test_set_current_linetype(doc):
    doc.linetypes.add("TEST", [0.0])
    appsettings.set_current_linetype(doc, "TEST")
    assert doc.header["$CELTYPE"] == "TEST"


def test_invalid_linetype_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_linetype(doc, "INVALID")


def test_set_current_lineweight(doc):
    appsettings.set_current_lineweight(doc, 50)
    assert doc.header["$CELWEIGHT"] == 50


def test_invalid_lineweight_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_lineweight(doc, 300)


def test_set_current_linetype_scale(doc):
    appsettings.set_current_linetype_scale(doc, 2.0)
    assert doc.header["$CELTSCALE"] == 2.0


def test_invalid_linetype_scale_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_linetype_scale(doc, 0)


def test_set_current_textstyle(doc):
    doc.styles.add("TEST", font="arial.ttf")
    appsettings.set_current_textstyle(doc, "TEST")
    assert doc.header["$TEXTSTYLE"] == "TEST"


def test_invalid_textstyle_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_textstyle(doc, "INVALID")


def test_set_current_dimstyle(doc):
    doc.dimstyles.add("TEST")
    appsettings.set_current_dimstyle(doc, "TEST")
    assert doc.header["$DIMSTYLE"] == "TEST"


def test_invalid_dimstyle_raises_exception(doc):
    with pytest.raises(ezdxf.DXFValueError):
        appsettings.set_current_dimstyle(doc, "INVALID")


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.math import BoundingBox, BoundingBox2d, Vec2, Vec3
from itertools import permutations


class TestBoundingBox:
    def test_init(self):
        bbox = BoundingBox([(0, 0, 0), (10, 10, 10)])
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

        bbox = BoundingBox([(10, 10, 10), (0, 0, 0)])
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

        bbox = BoundingBox([(7, -2, 9), (-1, 8, -3)])
        assert bbox.extmin == (-1, -2, -3)
        assert bbox.extmax == (7, 8, 9)

    def test_init_none(self):
        bbox = BoundingBox()
        assert bbox.is_empty is True
        assert bbox.has_data is False

        bbox.extend([(0, 0, 0), (10, 10, 10)])
        assert bbox.size == (10, 10, 10)
        assert bbox.is_empty is False
        assert bbox.has_data is True

    def test_init_empty_list(self):
        bbox = BoundingBox([])
        assert bbox.has_data is False

    @pytest.mark.parametrize(
        "v",
        [
            [],
            [(0, 0, 0)],
            [(1, 0, 0), (2, 0, 0)],
            [(1, 1, 0), (2, 2, 0)],
            [(1, 1, 1), (1, 1, 1)],
        ],
    )
    def test_is_empty(self, v):
        """Volume is 0."""
        assert BoundingBox(v).is_empty is True

    def test_not_is_empty(self):
        """Volume is not 0."""
        assert BoundingBox([(0, 0, 0), (-1, -1, -1)]).is_empty is False

    def test_init_with_with_empty_list(self):
        assert BoundingBox([]).is_empty is True

    def test_inside(self):
        bbox = BoundingBox([(0, 0, 0), (10, 10, 10)])
        assert bbox.inside((0, 0, 0)) is True
        assert bbox.inside((-1, 0, 0)) is False
        assert bbox.inside((0, -1, 0)) is False
        assert bbox.inside((0, 0, -1)) is False

        assert bbox.inside((5, 5, 5)) is True

        assert bbox.inside((10, 10, 10)) is True
        assert bbox.inside((11, 10, 10)) is False
        assert bbox.inside((10, 11, 10)) is False
        assert bbox.inside((10, 10, 11)) is False

    def test_all_inside(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox([(1, 1, 1), (9, 9, 9)])
        empty = BoundingBox()
        assert bbox1.all_inside(bbox2) is True
        assert bbox2.all_inside(bbox1) is False
        assert bbox1.all_inside(empty) is False
        assert empty.all_inside(bbox1) is False

    def test_any_inside(self):
        bbox1 = BoundingBox([(0, 0, 0), (7, 7, 7)])
        bbox2 = BoundingBox([(1, 1, 1), (9, 9, 9)])
        empty = BoundingBox()
        assert bbox1.any_inside(bbox2) is True
        assert bbox2.any_inside(bbox1) is True
        assert bbox1.any_inside(empty) is False
        assert empty.any_inside(bbox1) is False

    def test_has_overlap_accepts_2d_bounding_box(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox2d([(1, 1), (9, 9)])  # z-axis are 0
        assert bbox1.has_overlap(bbox2) is True

    def test_has_intersection_accepts_2d_bounding_box(self):
        bbox1 = BoundingBox([(-1, -1, -1), (10, 10, 10)])
        bbox2 = BoundingBox2d([(1, 1), (9, 9)])  # z-axis are 0
        assert bbox1.has_intersection(bbox2) is True

    def test_do_intersect_and_overlap(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox([(1, 1, 1), (9, 9, 9)])
        bbox3 = BoundingBox([(-1, -1, -1), (1.001, 1.001, 1.001)])
        for a, b in permutations([bbox1, bbox2, bbox3], 2):
            assert a.has_intersection(b) is True
            assert a.has_overlap(b) is True

    def test_do_not_intersect_or_overlap(self):
        bbox1 = BoundingBox([(0, 0, 0), (3, 3, 3)])
        bbox2 = BoundingBox([(4, 4, 4), (9, 9, 9)])
        bbox3 = BoundingBox([(-2, -2, -2), (-1, -1, -1)])
        for a, b in permutations([bbox1, bbox2, bbox3], 2):
            assert a.has_intersection(b) is False
            assert a.has_overlap(b) is False

    def test_do_not_intersect_or_overlap_empty(self):
        bbox = BoundingBox([(0, 0, 0), (3, 3, 3)])
        empty = BoundingBox()
        assert bbox.has_intersection(empty) is False
        assert bbox.has_overlap(empty) is False
        assert empty.has_intersection(bbox) is False
        assert empty.has_overlap(bbox) is False
        assert empty.has_intersection(empty) is False
        assert empty.has_overlap(empty) is False

    def test_crossing_2d_boxes(self):
        # bboxes do overlap, but do not contain corner points of the other bbox
        bbox1 = BoundingBox2d([(0, 1), (3, 2)])
        bbox2 = BoundingBox2d([(1, 0), (2, 3)])
        assert bbox1.has_intersection(bbox2) is True
        assert bbox1.has_overlap(bbox2) is True

    def test_crossing_3d_boxes(self):
        # bboxes do overlap, but do not contain corner points of the other bbox
        bbox1 = BoundingBox([(0, 1, 0), (3, 2, 1)])
        bbox2 = BoundingBox([(1, 0, 0), (2, 3, 1)])
        assert bbox1.has_intersection(bbox2) is True
        assert bbox1.has_overlap(bbox2) is True

    def test_touching_2d_boxes(self):
        bbox1 = BoundingBox2d([(0, 0), (1, 1)])
        bbox2 = BoundingBox2d([(1, 1), (2, 2)])
        bbox3 = BoundingBox2d([(-1, -1), (0, 0)])
        assert bbox1.has_intersection(bbox2) is False
        assert bbox1.has_overlap(bbox2) is True
        assert bbox2.has_intersection(bbox1) is False
        assert bbox2.has_overlap(bbox1) is True
        assert bbox1.has_intersection(bbox3) is False
        assert bbox1.has_overlap(bbox3) is True
        assert bbox3.has_intersection(bbox1) is False
        assert bbox3.has_overlap(bbox1) is True

    def test_touching_3d_boxes(self):
        bbox1 = BoundingBox([(0, 0, 0), (1, 1, 1)])
        bbox2 = BoundingBox([(1, 1, 1), (2, 2, 2)])
        bbox3 = BoundingBox([(-1, -1, -1), (0, 0, 0)])
        assert bbox1.has_intersection(bbox2) is False
        assert bbox1.has_overlap(bbox2) is True
        assert bbox2.has_intersection(bbox1) is False
        assert bbox2.has_overlap(bbox1) is True
        assert bbox1.has_intersection(bbox3) is False
        assert bbox1.has_overlap(bbox3) is True
        assert bbox3.has_intersection(bbox1) is False
        assert bbox3.has_overlap(bbox1) is True

    def test_extend(self):
        bbox = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox.extend([(5, 5, 5)])
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

        bbox.extend([(15, 16, 17)])
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (15, 16, 17)

        bbox.extend([(-15, -16, -17)])
        assert bbox.extmin == (-15, -16, -17)
        assert bbox.extmax == (15, 16, 17)

    def test_extend_by_empty_list(self):
        bbox = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox.extend([])
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

    def test_extend_by_bbox(self):
        bbox = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox([(5, 5, 5), (15, 16, 17)])
        bbox.extend(bbox2)
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (15, 16, 17)

    def test_size(self):
        bbox = BoundingBox([(-2, -2, -2), (8, 8, 8)])
        assert bbox.size == (10, 10, 10)

    def test_center(self):
        bbox = BoundingBox([(-1, -1, -1), (9, 9, 9)])
        assert bbox.center == (4, 4, 4)

    def test_union_of_two_bounding_boxes(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox([(5, 5, 5), (15, 16, 17)])
        bbox = bbox1.union(bbox2)
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (15, 16, 17)

    def test_union_bbox_with_empty_bbox(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox = bbox1.union(BoundingBox())
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

    def test_union_empty_bbox_with_bbox(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox = BoundingBox().union(bbox1)
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (10, 10, 10)

    def test_union_empty_bounding_boxes(self):
        bbox = BoundingBox().union(BoundingBox())
        assert bbox.has_data is False
        assert bbox.is_empty is True

    def test_union_different_bounding_boxes(self):
        bbox1 = BoundingBox([(0, 0, 0), (10, 10, 10)])
        bbox2 = BoundingBox2d([(5, 5), (15, 16)])
        bbox = bbox1.union(BoundingBox(bbox2))
        assert bbox.extmin == (0, 0, 0)
        assert bbox.extmax == (15, 16, 10)

    def test_rect_vertices_for_empty_bbox_raises_value_error(self):
        with pytest.raises(ValueError):
            BoundingBox().rect_vertices()

    def test_cube_vertices_for_empty_bbox_raises_value_error(self):
        with pytest.raises(ValueError):
            BoundingBox().cube_vertices()

    def test_rect_vertices_returns_vertices_in_counter_clockwise_order(self):
        bbox = BoundingBox([(0, 0, 0), (1, 2, 3)])
        assert bbox.rect_vertices() == Vec2.tuple([(0, 0), (1, 0), (1, 2), (0, 2)])

    def test_cube_vertices_returns_vertices_in_counter_clockwise_order(self):
        bbox = BoundingBox([(0, 0, 0), (1, 2, 3)])
        assert bbox.cube_vertices() == Vec3.tuple(
            [
                (0, 0, 0),  # bottom layer
                (1, 0, 0),
                (1, 2, 0),
                (0, 2, 0),
                (0, 0, 3),  # top layer
                (1, 0, 3),
                (1, 2, 3),
                (0, 2, 3),
            ]
        )

    def test_contains_other_bounding_box(self):
        box_a = BoundingBox([(0, 0, 0), (10, 10, 10)])
        box_b = BoundingBox([(1, 1, 1), (9, 9, 9)])
        box_c = BoundingBox([(1, 1, 1), (11, 11, 11)])
        assert box_a.contains(box_b) is True
        assert box_a.contains(box_a) is True  # self contained
        assert box_b.contains(box_a) is False
        assert box_a.contains(box_c) is False

    def test_growing_empty_bounding_box_does_nothing(self):
        box = BoundingBox()
        box.grow(1)
        assert box.has_data is False

    def test_grow_bounding_box(self):
        box = BoundingBox([(0, 0, 0), (1, 1, 1)])
        box.grow(1)
        assert box.extmin.isclose((-1, -1, -1))
        assert box.extmax.isclose((2, 2, 2))

    def test_shrinking_to_zero_or_below_raises_exception(self):
        box = BoundingBox([(0, 0, 0), (1, 1, 1)])
        with pytest.raises(ValueError):
            box.grow(-0.5)


class TestBoundingBox2d:
    def test_init(self):
        bbox = BoundingBox2d([(0, 0), (10, 10)])
        assert bbox.extmin == (0, 0)
        assert bbox.extmax == (10, 10)

        bbox = BoundingBox2d([(7, -2), (-1, 8)])
        assert bbox.extmin == (-1, -2)
        assert bbox.extmax == (7, 8)

    def test_init_with_with_empty_list(self):
        assert BoundingBox2d([]).is_empty is True

    def test_init_none(self):
        bbox = BoundingBox2d()
        assert bbox.has_data is False
        bbox.extend([(0, 0), (10, 10)])
        assert bbox.size == (10, 10)
        assert bbox.has_data is True

    @pytest.mark.parametrize(
        "v",
        [
            [],
            [(0, 0)],
            [(1, 0), (2, 0)],
            [(0, 1), (0, 2)],
            [(1, 1), (1, 1)],
        ],
    )
    def test_is_empty(self, v):
        """Area is 0."""
        assert BoundingBox2d(v).is_empty is True

    def test_not_is_empty(self):
        """Area is not 0."""
        assert BoundingBox2d([(0, 0), (-1, -1)]).is_empty is False

    def test_inside(self):
        bbox = BoundingBox2d([(0, 0), (10, 10)])
        assert bbox.inside((0, 0)) is True
        assert bbox.inside((-1, 0)) is False
        assert bbox.inside((0, -1)) is False

        assert bbox.inside((5, 5)) is True

        assert bbox.inside((10, 10)) is True
        assert bbox.inside((11, 10)) is False
        assert bbox.inside((10, 11)) is False

    def test_extend(self):
        bbox = BoundingBox2d([(0, 0), (10, 10)])
        bbox.extend([(5, 5)])
        assert bbox.extmin == (0, 0)
        assert bbox.extmax == (10, 10)

        bbox.extend([(15, 16)])
        assert bbox.extmin == (0, 0)
        assert bbox.extmax == (15, 16)

        bbox.extend([(-15, -16)])
        assert bbox.extmin == (-15, -16)
        assert bbox.extmax == (15, 16)

    def test_size(self):
        bbox = BoundingBox2d([(-2, -2), (8, 8)])
        assert bbox.size == (10, 10)

    def test_center(self):
        bbox = BoundingBox2d([(-1, -1), (9, 9)])
        assert bbox.center == (4, 4)

    def test_rect_vertices_for_empty_bbox_raises_value_error(self):
        with pytest.raises(ValueError):
            BoundingBox2d().rect_vertices()

    def test_rect_vertices_returns_vertices_in_counter_clockwise_order(self):
        bbox = BoundingBox2d([(0, 0, 0), (1, 2, 3)])
        assert bbox.rect_vertices() == Vec2.tuple([(0, 0), (1, 0), (1, 2), (0, 2)])

    def test_contains_other_bounding_box(self):
        box_a = BoundingBox2d([(0, 0), (10, 10)])
        box_b = BoundingBox2d([(1, 1), (9, 9)])
        box_c = BoundingBox2d([(1, 1), (11, 11)])
        assert box_a.contains(box_b) is True
        assert box_a.contains(box_a) is True  # self contained!
        assert box_b.contains(box_a) is False
        assert box_a.contains(box_c) is False

    def test_2d_box_contains_3d_box(self):
        # casting is required since v1.2 for valid type-checking
        box_a = BoundingBox2d(
            [(0, 0), (10, 10)]
        )  # this is flat-land, z-axis does not exist
        box_b = BoundingBox([(1, 1, 1), (9, 9, 9)])
        assert box_a.contains(BoundingBox2d(box_b)) is True, "z-axis should be ignored"

    def test_3d_box_contains_2d_box(self):
        # casting is required since v1.2 for valid type-checking
        box_a = BoundingBox2d([(1, 1), (9, 9)])  # lives in the xy-plane, z-axis is 0
        box_b = BoundingBox([(0, 0, 0), (10, 10, 10)])
        assert box_b.contains(BoundingBox(box_a)) is True, "xy-plane is included"
        box_c = BoundingBox([(0, 0, 1), (10, 10, 10)])
        assert box_c.contains(BoundingBox(box_a)) is False, "xy-plane is not included"

    def test_grow_bounding_box(self):
        box = BoundingBox2d([(0, 0), (1, 1)])
        box.grow(1)
        assert box.extmin.isclose((-1, -1))
        assert box.extmax.isclose((2, 2))

    def test_has_overlap_accepts_3d_bounding_box(self):
        bbox1 = BoundingBox([(0, 0, 10), (10, 10, 10)])  # z-axis are ignored
        bbox2 = BoundingBox2d([(1, 1), (9, 9)])
        assert bbox2.has_overlap(bbox1) is True

    def test_has_intersection_accepts_3d_bounding_box(self):
        bbox1 = BoundingBox([(0, 0, 10), (10, 10, 10)])  # z-axis are ignored
        bbox2 = BoundingBox2d([(1, 1), (9, 9)])
        assert bbox2.has_intersection(bbox1) is True


class Test2DIntersection:
    def test_accept_3d_box(self):
        b1 = BoundingBox2d([(1, 1), (9, 9)])
        b2 = BoundingBox([(0, 0, 10), (10, 10, 10)])  # z-axis are ignored
        b = b1.intersection(b2)
        assert b.size.isclose((8, 8))

    def test_empty_box_intersection(self):
        b = BoundingBox2d().intersection(BoundingBox2d())
        assert b.is_empty is True

    def test_no_intersection(self):
        b1 = BoundingBox2d([(0, 0), (1, 1)])
        b2 = BoundingBox2d([(2, 2), (3, 3)])
        b = b1.intersection(b2)
        assert b.is_empty is True

    def test_touches_at_one_corner(self):
        b1 = BoundingBox2d([(0, 0), (1, 1)])
        b2 = BoundingBox2d([(1, 1), (2, 2)])
        b = b1.intersection(b2)
        assert b.is_empty is True

    def test_half_intersection(self):
        b1 = BoundingBox2d([(0, 0), (2, 2)])
        b2 = BoundingBox2d([(1, 1), (3, 3)])
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1))
        assert b.extmin.isclose((1, 1))
        assert b.extmax.isclose((2, 2))

    @pytest.mark.parametrize(
        "v1, v2",
        [
            ([(0, 0), (2, 2)], [(1, 1), (3, 3)]),
            ([(-1, -1), (1, 1)], [(0, 0), (2, 2)]),
            ([(-2, -2), (0, 0)], [(-3, -3), (-1, -1)]),
        ],
    )
    def test_multiple_intersections(self, v1, v2):
        b1 = BoundingBox2d(v1)
        b2 = BoundingBox2d(v2)
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1))

    def test_full_intersection(self):
        b1 = BoundingBox2d([(0, 0), (2, 2)])
        b = b1.intersection(b1)
        assert b.size.isclose((2, 2))
        assert b.extmin.isclose((0, 0))
        assert b.extmax.isclose((2, 2))

    def test_smaller_inside_bigger_intersection(self):
        b1 = BoundingBox2d([(0, 0), (3, 3)])
        b2 = BoundingBox2d([(1, 1), (2, 2)])
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1))
        assert b.extmin.isclose((1, 1))
        assert b.extmax.isclose((2, 2))

    def test_intersection_box_without_an_area_is_empty(self):
        b1 = BoundingBox2d([(0, 0), (0, 3)])  # has no area
        b2 = BoundingBox2d([(-1, 1), (1, 1)])  # has no area

        b = b1.intersection(b2)
        assert b.is_empty is True  # intersection has no area


class Test3DIntersection:
    def test_accept_3d_box(self):
        b1 = BoundingBox([(0, 0, -1), (10, 10, 10)])
        b2 = BoundingBox2d([(1, 1), (9, 9)])  # z-axis is 0
        b = b1.intersection(b2)
        assert b.is_empty is True  # has no volume!

    def test_empty_box_intersection(self):
        b = BoundingBox().intersection(BoundingBox())
        assert b.is_empty is True

    def test_no_intersection(self):
        b1 = BoundingBox([(0, 0, 0), (1, 1, 1)])
        b2 = BoundingBox([(2, 2, 2), (3, 3, 3)])
        b = b1.intersection(b2)
        assert b.is_empty is True

    def test_touches_at_one_corner(self):
        b1 = BoundingBox([(0, 0, 0), (1, 1, 1)])
        b2 = BoundingBox([(1, 1, 1), (2, 2, 2)])
        b = b1.intersection(b2)
        assert b.is_empty is True

    def test_half_intersection(self):
        b1 = BoundingBox([(0, 0, 0), (2, 2, 2)])
        b2 = BoundingBox([(1, 1, 1), (3, 3, 3)])
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1, 1))
        assert b.extmin.isclose((1, 1, 1))
        assert b.extmax.isclose((2, 2, 2))

    @pytest.mark.parametrize(
        "v1, v2",
        [
            ([(0, 0, 0), (2, 2, 2)], [(1, 1, 1), (3, 3, 3)]),
            ([(-1, -1, -1), (1, 1, 1)], [(0, 0, 0), (2, 2, 2)]),
            ([(-2, -2, -2), (0, 0, 0)], [(-3, -3, -3), (-1, -1, -1)]),
        ],
    )
    def test_multiple_intersections(self, v1, v2):
        b1 = BoundingBox(v1)
        b2 = BoundingBox(v2)
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1, 1))

    def test_full_intersection(self):
        b1 = BoundingBox([(0, 0, 0), (2, 2, 2)])
        b = b1.intersection(b1)
        assert b.size.isclose((2, 2, 2))
        assert b.extmin.isclose((0, 0, 0))
        assert b.extmax.isclose((2, 2, 2))

    def test_smaller_inside_bigger_intersection(self):
        b1 = BoundingBox([(0, 0, 0), (3, 3, 3)])
        b2 = BoundingBox([(1, 1, 1), (2, 2, 2)])
        b = b1.intersection(b2)
        assert b.size.isclose((1, 1, 1))
        assert b.extmin.isclose((1, 1, 1))
        assert b.extmax.isclose((2, 2, 2))

    def test_intersection_box_without_a_volume_is_empty(self):
        b1 = BoundingBox2d([(0, 0, 0), (0, 3, 0)])  # has no volume
        b2 = BoundingBox2d([(-1, 1, 0), (1, 1, 0)])  # has no volume
        b = b1.intersection(b2)
        assert b.is_empty is True  # intersection has no volume



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from typing import Any
import pytest

from ezdxf.math import Vec2
from ezdxf.addons.drawing.backend import BackendProperties, BkPoints2d, BkPath2d
from ezdxf.addons.drawing import json
from ezdxf import path


def test_properties():
    backend = json.GeoJSONBackend()
    properties = BackendProperties(color="#ff0000", lineweight=0.25, layer="One")
    backend.draw_point(Vec2(13, 17), properties)
    result = backend.get_json_data()
    assert result["type"] == "FeatureCollection"
    point = result["features"][0]
    assert point["type"] == "Feature"
    assert point["properties"]["color"] == "#ff0000"
    assert point["properties"]["stroke-width"] == 0.25
    assert point["properties"]["layer"] == "One"
    assert point["geometry"] == {"type": "Point", "coordinates": [13, 17]}


def no_properties(color: str, stroke_width: float, layer: str) -> dict[str, Any]:
    return {}


class TestGeoJSONBackendNoProperties:
    @pytest.fixture
    def backend(self):
        return json.GeoJSONBackend(properties_maker=no_properties)

    @pytest.fixture
    def properties(self):
        return BackendProperties(color="#ff0000", lineweight=0.25, layer="One")

    def test_draw_point(
        self, backend: json.GeoJSONBackend, properties: BackendProperties
    ):
        backend.draw_point(Vec2(13, 17), properties)
        result = backend.get_json_data()
        assert result["type"] == "GeometryCollection"
        point = result["geometries"][0]
        assert point["type"] == "Point"
        assert point["coordinates"] == [13, 17]

    def test_draw_solid_lines(
        self, backend: json.GeoJSONBackend, properties: BackendProperties
    ):
        backend.draw_solid_lines(
            [
                (Vec2(0, 0), Vec2(1, 0)),
                (Vec2(2, 0), Vec2(3, 0)),
                (Vec2(4, 0), Vec2(5, 0)),
            ],
            properties,
        )
        result = backend.get_json_data()
        lines = result["geometries"][0]
        assert lines["type"] == "MultiLineString"
        assert lines["coordinates"] == [
            ((0, 0), (1, 0)),
            ((2, 0), (3, 0)),
            ((4, 0), (5, 0)),
        ]

    def test_draw_path(
        self, backend: json.GeoJSONBackend, properties: BackendProperties
    ):
        p = path.Path((2, 2))
        p.line_to((4, 2))
        p.curve3_to((8, 2), (6, 3))
        p.curve4_to((11, 2), (9, 3), (10, 3))

        backend.draw_path(BkPath2d(p), properties)
        result = backend.get_json_data()
        json_path = result["geometries"][0]
        assert json_path["type"] == "LineString"
        coordinates = json_path["coordinates"]

        assert coordinates[0] == (2, 2)
        assert coordinates[-1] == (11, 2)

    def test_draw_filled_polygon(
        self, backend: json.GeoJSONBackend, properties: BackendProperties
    ):
        polygon = BkPoints2d(Vec2.list([(0, 0), (2, 0), (1, 1)]))
        backend.draw_filled_polygon(polygon, properties)
        result = backend.get_json_data()
        json_polygon = result["geometries"][0]
        assert json_polygon["type"] == "Polygon"
        assert json_polygon["coordinates"] == [[(0, 0), (2, 0), (1, 1), (0, 0)]]


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# Created 2019-02-13
import pytest
import ezdxf
from ezdxf.document import Drawing
from ezdxf.entities import factory, DXFEntity
from ezdxf.entities.xdict import ExtensionDict


@pytest.fixture(scope="module")
def doc() -> Drawing:
    return ezdxf.new()


@pytest.fixture()
def entity(doc):
    msp = doc.modelspace()
    return msp.add_line((0, 0), (1, 1))


def test_new_extension_dict(doc, entity):
    factory.bind(entity, doc)
    assert entity.has_extension_dict is False
    xdict = entity.new_extension_dict()
    assert xdict.dictionary.dxftype() == "DICTIONARY"
    assert len(xdict.dictionary) == 0

    placeholder = xdict.add_placeholder("TEST")
    assert len(xdict.dictionary) == 1
    assert placeholder.dxf.owner == xdict.dictionary.dxf.handle
    assert "TEST" in xdict.dictionary


def test_direct_interface(doc, entity):
    factory.bind(entity, doc)
    xdict = entity.new_extension_dict()
    placeholder = xdict.add_placeholder("TEST")
    assert "TEST" in xdict
    placeholder2 = xdict["TEST"]
    assert placeholder is placeholder2
    xdict["TEST2"] = placeholder2


def test_get_returns_none_for_non_existing_entries(doc):
    xdict = ExtensionDict(doc.objects.add_dictionary())
    assert xdict.get("ABC") is None


def test_supports_handle_property(doc):
    xdict = ExtensionDict.new("ABBA", doc)
    assert isinstance(xdict.handle, str)


def test_line_new_extension_dict(doc):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    assert entity.has_extension_dict is False
    xdict = entity.new_extension_dict()
    dxf_dict = xdict.dictionary
    assert dxf_dict.dxftype() == "DICTIONARY"
    assert dxf_dict.dxf.owner == entity.dxf.handle
    assert (
        entity.has_app_data("{ACAD_XDICTIONARY") is False
    ), "extension dictionary is a separated storage"
    assert entity.has_extension_dict is True

    xdict2 = entity.get_extension_dict()
    dxf_dict2 = xdict2.dictionary
    assert dxf_dict.dxf.handle == dxf_dict2.dxf.handle


def test_del_entity_with_ext_dict(doc):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    xdict = entity.new_extension_dict()

    objects = doc.objects
    assert xdict.dictionary in objects
    store_xdict = xdict.dictionary
    msp.delete_entity(entity)
    doc.objects.purge()
    assert xdict.is_alive is False
    assert store_xdict not in objects


def test_discard_existing_extension_dict(doc):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    xdict = entity.new_extension_dict()
    dictionary = xdict.dictionary
    placeholder = xdict.add_placeholder("TEST")
    entity.discard_extension_dict()
    assert placeholder.is_alive is False
    assert dictionary.is_alive is False
    assert xdict.is_alive is False
    assert entity.has_extension_dict is False


def test_discard_non_existing_extension_dict_without_exception(doc: Drawing):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    entity.discard_extension_dict()
    assert entity.has_extension_dict is False


def test_discard_empty_extension(doc: Drawing):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    entity.discard_empty_extension_dict()
    assert entity.has_extension_dict is False

def test_do_not_discard_non_empty_extension(doc: Drawing):
    msp = doc.modelspace()
    entity = msp.add_line((0, 0), (10, 0))
    xdict = entity.new_extension_dict()
    key = "PLACEHOLDER_KEY"
    xdict.add_placeholder(key)
    entity.discard_empty_extension_dict()
    assert entity.has_extension_dict is True
    assert key in xdict


def test_multiple_destroy_calls(doc, entity):
    xdict = ExtensionDict.new("ABBA", doc)
    xdict.destroy()
    xdict.destroy(), "2nd call should not raise an exception"
    assert xdict.is_alive is False
    assert entity.has_extension_dict is False


def test_link_dxf_object_to_extension_dict(doc):
    from ezdxf.entities import DXFObject

    xdict = ExtensionDict.new("ABBA", doc)
    owner = xdict.handle
    obj = DXFObject.new(handle="FEFE")
    xdict.link_dxf_object("MyEntry", obj)
    assert "MyEntry" in xdict
    assert obj.dxf.owner == owner


def test_add_new_dictionary_to_xdict(doc):
    # issue #751
    msp = doc.modelspace()
    line = msp.add_line((0, 0), (1, 0))
    xdict = line.new_extension_dict()
    new_dict = xdict.add_dictionary("TEST")
    assert new_dict.dxf.owner == xdict.handle


if __name__ == '__main__':
    pytest.main([__file__])


# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# created 2019-02-15
import pytest

from ezdxf.entities.image import RasterVariables
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

RASTERVARIABLES = """0
RASTERVARIABLES
5
0
102
{ACAD_REACTORS
330
0
102
}
330
0
100
AcDbRasterVariables
90
0
70
0
71
1
72
3
"""


@pytest.fixture
def entity():
    return RasterVariables.from_text(RASTERVARIABLES)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "RASTERVARIABLES" in ENTITY_CLASSES


def test_default_init():
    entity = RasterVariables()
    assert entity.dxftype() == "RASTERVARIABLES"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = RasterVariables.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.class_version == 0
    assert entity.dxf.frame == 0
    assert entity.dxf.quality == 1
    assert entity.dxf.units == 3


def test_load_from_text(entity):
    assert entity.dxf.class_version == 0
    assert entity.dxf.frame == 0
    assert entity.dxf.quality == 1
    assert entity.dxf.units == 3


def test_write_dxf():
    entity = RasterVariables.from_text(RASTERVARIABLES)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(RASTERVARIABLES)
    assert result == expected



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
from ezdxf.math import (
    Vec3,
    estimate_tangents,
    local_cubic_bspline_interpolation,
)
from ezdxf.math.bspline import local_cubic_bspline_interpolation_from_tangents


POINTS1 = [(1, 1), (2, 4), (4, 1), (7, 6)]
POINTS2 = [(1, 1), (2, 4), (4, 1), (7, 6), (5, 8), (3, 3), (1, 7)]


def test_estimate_tangents_3p():
    tangents = estimate_tangents(Vec3.list(POINTS1), method="3-points")
    assert len(tangents) == 4


def test_estimate_tangents_5p():
    tangents = estimate_tangents(Vec3.list(POINTS1), method="5-points")
    assert len(tangents) == 4


def test_local_cubic_bspline_interpolation_from_tangents():
    points = Vec3.list(POINTS1)
    tangents = estimate_tangents(points)
    control_points, knots = local_cubic_bspline_interpolation_from_tangents(
        points, tangents
    )
    assert len(control_points) == 8
    assert len(knots) == 8 + 4  # count + order


def test_local_cubic_bspline_interpolation():
    bspline = local_cubic_bspline_interpolation(POINTS1, method="5-points")
    assert bspline.degree == 3
    assert len(bspline.control_points) == 8
    assert len(bspline.knots()) == 8 + 4  # count + order



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Optional, Iterator, TYPE_CHECKING, overload
import math
from ezdxf.math import Vec2, UVec, Matrix44


__all__ = ["Shape2d"]

if TYPE_CHECKING:
    from ezdxf.math import BoundingBox2d


class Shape2d:
    """Construction tools for 2D shapes.

    A 2D geometry object as list of :class:`Vec2` objects, vertices can be
    moved, rotated and scaled.

    Args:
        vertices: iterable of :class:`Vec2` compatible objects.

    """

    def __init__(self, vertices: Optional[Iterable[UVec]] = None):
        from ezdxf.npshapes import NumpyPoints2d

        _vec2: Iterator[Vec2] | None = None
        if vertices is not None:
            _vec2 = Vec2.generate(vertices)
        self.np_vertices = NumpyPoints2d(_vec2)

    @property
    def vertices(self) -> list[Vec2]:
        return self.np_vertices.vertices()

    @property
    def bounding_box(self) -> BoundingBox2d:
        """Returns the bounding box of the shape."""
        return self.np_vertices.bbox()

    def copy(self) -> Shape2d:
        clone = self.__class__()
        clone.np_vertices = self.np_vertices.clone()
        return clone

    __copy__ = copy

    def translate(self, vector: UVec) -> None:
        """Translate shape about `vector`."""
        offset = Vec2(vector)
        self.np_vertices.transform_inplace(Matrix44.translate(offset.x, offset.y, 0))

    def scale(self, sx: float = 1.0, sy: float = 1.0) -> None:
        """Scale shape about `sx` in x-axis and `sy` in y-axis."""
        self.np_vertices.transform_inplace(Matrix44.scale(sx, sy, 1))

    def scale_uniform(self, scale: float) -> None:
        """Scale shape uniform about `scale` in x- and y-axis."""
        self.np_vertices.transform_inplace(Matrix44.scale(scale, scale, 1))

    def rotate(self, angle: float, center: Optional[UVec] = None) -> None:
        """Rotate shape around rotation `center` about `angle` in degrees."""
        self.rotate_rad(math.radians(angle), center)

    def rotate_rad(self, angle: float, center: Optional[UVec] = None) -> None:
        """Rotate shape around rotation `center` about `angle` in radians."""
        m = Matrix44.z_rotate(angle)
        if center is not None:
            p = Vec2(center)
            m = (
                Matrix44.translate(-p.x, -p.y, 0)
                @ m
                @ Matrix44.translate(p.x, p.y, 0)
            )
        self.np_vertices.transform_inplace(m)

    def offset(self, offset: float, closed: bool = False) -> Shape2d:
        """Returns a new offset shape, for more information see also
        :func:`ezdxf.math.offset_vertices_2d` function.

        Args:
            offset: line offset perpendicular to direction of shape segments
                defined by vertices order, offset > ``0`` is 'left' of line
                segment, offset < ``0`` is 'right' of line segment
            closed: ``True`` to handle as closed shape

        """
        from ezdxf.math.offset2d import offset_vertices_2d

        return self.__class__(
            offset_vertices_2d(
                self.np_vertices.vertices(), offset=offset, closed=closed
            )
        )

    def convex_hull(self) -> Shape2d:
        """Returns convex hull as new shape."""
        from ezdxf.math.construct2d import convex_hull_2d

        return self.__class__(convex_hull_2d(self.np_vertices.vertices()))

    # Sequence interface
    def __len__(self) -> int:
        """Returns `count` of vertices."""
        return len(self.np_vertices)

    @overload
    def __getitem__(self, item: int) -> Vec2: ...
    @overload
    def __getitem__(self, item: slice) -> list[Vec2]: ...
    def __getitem__(self, item: int | slice) -> Vec2|list[Vec2]:
        """Get vertex by index `item`, supports ``list`` like slicing."""
        np_vertices = self.np_vertices.np_vertices()
        if isinstance(item, int):
            return Vec2(np_vertices[item])
        else:
            return Vec2.list(np_vertices[item])

    def append(self, vertex: UVec) -> None:
        """Append single `vertex`.

        Args:
             vertex: vertex as :class:`Vec2` compatible object

        """
        self.extend((vertex,))

    def extend(self, vertices: Iterable[UVec]) -> None:
        """Append multiple `vertices`.

        Args:
             vertices: iterable of vertices as :class:`Vec2` compatible objects

        """
        from ezdxf.npshapes import NumpyPoints2d

        new_vertices = self.np_vertices.vertices() + Vec2.list(vertices)
        self.np_vertices = NumpyPoints2d(new_vertices)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.lldxf import const
from ezdxf.math import Vec3
from ezdxf.lldxf.tags import Tags

# noinspection PyProtectedMember
from ezdxf.urecord import (
    parse_items,
    compile_user_record,
    parse_binary_data,
    UserRecord,
    BinaryRecord,
)


class TestFlatRecord:
    @pytest.fixture
    def tags(self):
        return Tags.from_text(FLAT_RECORD)

    def test_parse_all(self, tags):
        data = parse_items(tags)
        assert len(data) == 4

    def test_parse_str(self, tags):
        data = parse_items(tags)[0]
        assert data == "MyString"
        assert type(data) == str

    def test_parse_float(self, tags):
        data = parse_items(tags)[1]
        assert data == 3.1415
        assert type(data) is float

    def test_parse_int(self, tags):
        data = parse_items(tags)[2]
        assert data == 255
        assert type(data) is int

    def test_parse_vec3(self, tags):
        data = parse_items(tags)[3]
        assert data == (7, 8, 9)
        assert type(data) is Vec3


def test_top_level_list():
    tags = Tags.from_text(LIST1)
    data = parse_items(tags)
    assert data == ["MyString", ["ListItem1"], "Tail"]


def test_nested_list_inside_list():
    tags = Tags.from_text(LIST2)
    data = parse_items(tags)
    assert data == ["MyString", ["ListItem1", ["ListItem2"]], "Tail"]


def test_top_level_dict():
    tags = Tags.from_text(DICT1)
    data = parse_items(tags)
    assert data == ["MyString", {"Key1": "Value1"}, "Tail"]


def test_nested_dict_as_dict_value():
    tags = Tags.from_text(DICT2)
    data = parse_items(tags)
    assert data == ["MyString", {"Key1": {"Key2": "Value2"}}, "Tail"]


def test_nested_list_as_dict_value():
    tags = Tags.from_text(DICT_LIST)
    data = parse_items(tags)
    assert data == ["MyString", {"Key": ["ListItem"]}, "Tail"]


def test_nested_dict_inside_list():
    tags = Tags.from_text(LIST_DICT)
    data = parse_items(tags)
    assert data == [
        "MyString",
        ["ListItem1", {"Key": "Value"}, "ListItem2"],
        "Tail",
    ]


def test_missing_open_tag_raises_dxf_structure_error():
    tags = Tags.from_text("1\nListItem1\n302\n]")
    with pytest.raises(const.DXFStructureError):
        parse_items(tags)


def test_missing_close_tag_raises_dxf_structure_error():
    tags = Tags.from_text("302\n[\n1\nListItem1")
    with pytest.raises(const.DXFStructureError):
        parse_items(tags)


def test_invalid_group_code_raises_value_error():
    tags = Tags.from_text("5\ninvalid\n")
    with pytest.raises(const.DXFValueError):
        parse_items(tags)


@pytest.mark.parametrize("char", ["\n", "\r"])
def test_invalid_line_break_characters_raise_exception(char):
    with pytest.raises(const.DXFValueError):
        compile_user_record("TEST", [f"{char}"])


def test_too_long_string_raise_exception():
    # max. str length is 2049 - DXF R2000 limit for group codes 0-9
    with pytest.raises(const.DXFValueError):
        compile_user_record("TEST", ["0123456789" * 205])


class TestCompileData:
    def test_compile_empty_data(self):
        tags = compile_user_record("TEST", [])
        assert tags[0] == (2, "TEST")
        assert len(tags) == 1

    @pytest.mark.parametrize(
        "value",
        [
            "MyString",
            257,
            3.1415,
            Vec3(5, 6, 7),
        ],
        ids=["str", "int", "float", "Vec3"],
    )
    def test_compile_simple_types(self, value):
        tags = compile_user_record("TEST", [value])
        assert parse_items(tags[1:]) == [value]

    @pytest.mark.parametrize(
        "struct",
        [
            [1, 2, 3],  # simple flat list
            [{"key": "value"}],  # top level structure has to be a list!
            ["head", [1, 2, 3], "tail"],  # nested list
            ["head", {"key": "value"}, "tail"],  # nested dict
            ["head", {"key": [1, 2, 3]}, "tail"],  # list as dict value
            ["head", [1, 2, {"key": "value"}], "tail"],  # dict inside a list
        ],
        ids=[
            "flat list",
            "flat dict",
            "nested list",
            "nested dict",
            "list as dict value",
            "dict inside a list",
        ],
    )
    def test_compile_complex_structures(self, struct):
        tags = compile_user_record("TEST", struct)
        assert parse_items(tags[1:]) == struct


class TestUserRecord:
    def test_required_final_commit_to_store_data_in_xrecord(self):
        user_record = UserRecord(name="MyRecord")
        user_record.data.extend(["str", 1, 3.1415])
        assert len(user_record.xrecord.tags) == 0

        # calling commit() stores the data in the xrecord
        user_record.commit()
        assert user_record.xrecord.tags == [
            (2, "MyRecord"),
            (1, "str"),
            (90, 1),
            (40, 3.1415),
        ]

    def test_works_as_context_manager(self):
        with UserRecord(name="MyRecord") as user_record:
            user_record.data.extend(["str", 1, 3.1415])
            # calls commit() at exit

        assert user_record.xrecord.tags == [
            (2, "MyRecord"),
            (1, "str"),
            (90, 1),
            (40, 3.1415),
        ]

    def test_str(self):
        with UserRecord(name="MyRecord") as user_record:
            user_record.data.extend(["str", 1, 3.1415])
        assert str(user_record) == "['str', 1, 3.1415]"


def test_parse_binary_data():
    assert (
        parse_binary_data(Tags.from_text(BINARY_DATA))
        == b"0123456789\xab\xba" * 2
    )


class TestBinaryRecord:
    def test_required_final_commit_to_store_data_in_xrecord(self):
        user_record = BinaryRecord()
        user_record.data = b"\xfe\xfe"
        assert len(user_record.xrecord.tags) == 0
        # calling commit() stores the data in the xrecord
        user_record.commit()
        assert user_record.xrecord.tags == [(160, 2), (310, b"\xfe\xfe")]

    def test_works_as_context_manager(self):
        with BinaryRecord() as user_record:
            user_record.data = b"\xfe\xfe"
            # calls commit() at exit
        assert user_record.xrecord.tags == [(160, 2), (310, b"\xfe\xfe")]

    def test_stores_line_endings(self):
        with BinaryRecord() as user_record:
            user_record.data = b"\r\n"
        assert user_record.xrecord.tags == [(160, 2), (310, b"\r\n")]

    def test_str(self):
        record = BinaryRecord()
        record.data = b"\xfe\xfe"
        assert str(record) == "FEFE"


FLAT_RECORD = """1
MyString
40
3.1415
90
255
10
7
20
8
30
9
"""

LIST1 = """1
MyString
302
[
1
ListItem1
302
]
1
Tail
"""

LIST2 = """1
MyString
302
[
1
ListItem1
302
[
1
ListItem2
302
]
302
]
1
Tail
"""

DICT1 = """1
MyString
302
{
1
Key1
1
Value1
302
}
1
Tail
"""

DICT2 = """1
MyString
302
{
1
Key1
302
{
1
Key2
1
Value2
302
}
302
}
1
Tail
"""

DICT_LIST = """1
MyString
302
{
1
Key
302
[
1
ListItem
302
]
302
}
1
Tail
"""

LIST_DICT = """1
MyString
302
[
1
ListItem1
302
{
1
Key
1
Value
302
}
1
ListItem2
302
]
1
Tail
"""

BINARY_DATA = """160
24
310
30313233343536373839ABBA
310
30313233343536373839ABBA
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2014-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Any, Sequence, Union, overload, Optional
from array import array
import struct
from binascii import unhexlify, hexlify
from codecs import decode

Bytes = Union[bytes, bytearray, memoryview]


def hex_strings_to_bytes(data: Iterable[str]) -> bytes:
    """Returns multiple hex strings `data` as bytes."""
    byte_array = array("B")
    for hexstr in data:
        byte_array.extend(unhexlify(hexstr))
    return byte_array.tobytes()


def bytes_to_hexstr(data: bytes) -> str:
    """Returns `data` bytes as plain hex string."""
    return hexlify(data).upper().decode()


NULL_NULL = b"\x00\x00"


class EndOfBufferError(EOFError):
    pass


class ByteStream:
    """Process little endian binary data organized as bytes, data is padded to
    4 byte boundaries by default.
    """

    # Created for Proxy Entity Graphic decoding
    def __init__(self, buffer: Bytes, align: int = 4):
        self.buffer = memoryview(buffer)
        self.index: int = 0
        self._align: int = align

    @property
    def has_data(self) -> bool:
        return self.index < len(self.buffer)

    def align(self, index: int) -> int:
        modulo = index % self._align
        return index + self._align - modulo if modulo else index

    def read_struct(self, fmt: str) -> Any:
        """Read data defined by a struct format string. Insert little endian
        format character '<' as first character, if machine has native big
        endian byte order.
        """
        if not self.has_data:
            raise EndOfBufferError("Unexpected end of buffer.")

        result = struct.unpack_from(fmt, self.buffer, offset=self.index)
        self.index = self.align(self.index + struct.calcsize(fmt))
        return result

    def read_float(self) -> float:
        return self.read_struct("<d")[0]

    def read_long(self) -> int:
        return self.read_struct("<L")[0]

    def read_signed_long(self) -> int:
        return self.read_struct("<l")[0]

    def read_vertex(self) -> Sequence[float]:
        return self.read_struct("<3d")

    def read_padded_string(self, encoding: str = "utf_8") -> str:
        """PS: Padded String. This is a string, terminated with a zero byte.
        The file’s text encoding (code page) is used to encode/decode the bytes
        into a string.
        """
        buffer = self.buffer
        for end_index in range(self.index, len(buffer)):
            if buffer[end_index] == 0:
                start_index = self.index
                self.index = self.align(end_index + 1)
                # noinspection PyTypeChecker
                return decode(buffer[start_index:end_index], encoding=encoding)
        raise EndOfBufferError(
            "Unexpected end of buffer, did not detect terminating zero byte."
        )

    def read_padded_unicode_string(self) -> str:
        """PUS: Padded Unicode String. The bytes are encoded using Unicode
        encoding. The bytes consist of byte pairs and the string is terminated
        by 2 zero bytes.
        """
        buffer = self.buffer
        for end_index in range(self.index, len(buffer), 2):
            if buffer[end_index : end_index + 2] == NULL_NULL:
                start_index = self.index
                self.index = self.align(end_index + 2)
                # noinspection PyTypeChecker
                return decode(
                    buffer[start_index:end_index], encoding="utf_16_le"
                )
        raise EndOfBufferError(
            "Unexpected end of buffer, did not detect terminating zero bytes."
        )


class BitStream:
    """Process little endian binary data organized as bit stream."""

    # Created for Proxy Entity Graphic decoding and DWG bit stream decoding
    def __init__(
        self,
        buffer: Bytes,
        dxfversion: str = "AC1015",
        encoding: str = "cp1252",
    ):
        self.buffer = memoryview(buffer)
        self.bit_index: int = 0
        self.dxfversion = dxfversion
        self.encoding = encoding

    @property
    def has_data(self) -> bool:
        return self.bit_index >> 3 < len(self.buffer)

    def align(self, count: int) -> None:
        """Align to byte border."""
        byte_index = (self.bit_index >> 3) + bool(self.bit_index & 7)
        modulo = byte_index % count
        if modulo:
            byte_index += count - modulo
        self.bit_index = byte_index << 3

    def skip(self, count: int) -> None:
        """Skip `count` bits."""
        self.bit_index += count

    def read_bit(self) -> int:
        """Read one bit from buffer."""
        index = self.bit_index
        self.bit_index += 1
        try:
            return 1 if self.buffer[index >> 3] & (0x80 >> (index & 7)) else 0
        except IndexError:
            raise EndOfBufferError("Unexpected end of buffer.")

    def read_bits(self, count) -> int:
        """Read `count` bits from buffer."""
        index = self.bit_index
        buffer = self.buffer
        # index of next bit after reading `count` bits
        next_bit_index = index + count

        if (next_bit_index - 1) >> 3 > len(buffer):
            # not enough data to read all bits
            raise EndOfBufferError("Unexpected end of buffer.")
        self.bit_index = next_bit_index

        test_bit = 0x80 >> (index & 7)
        test_byte_index = index >> 3
        value = 0
        test_byte = buffer[test_byte_index]
        while count > 0:
            value <<= 1
            if test_byte & test_bit:
                value |= 1
            count -= 1
            test_bit >>= 1
            if not test_bit and count:
                test_bit = 0x80
                test_byte_index += 1
                test_byte = buffer[test_byte_index]
        return value

    def read_unsigned_byte(self) -> int:
        """Read an unsigned byte (8 bit) from buffer."""
        return self.read_bits(8)

    def read_signed_byte(self) -> int:
        """Read a signed byte (8 bit) from buffer."""
        value = self.read_bits(8)
        if value & 0x80:
            # 2er complement
            return -((~value & 0xFF) + 1)
        else:
            return value

    def read_aligned_bytes(self, count: int) -> Sequence[int]:
        buffer = self.buffer
        start_index = self.bit_index >> 3
        end_index = start_index + count
        if end_index <= len(buffer):
            self.bit_index += count << 3
            return buffer[start_index:end_index]
        else:
            raise EndOfBufferError("Unexpected end of buffer.")

    def read_unsigned_short(self) -> int:
        """Read an unsigned short (16 bit) from buffer."""
        if self.bit_index & 7:
            s1 = self.read_bits(8)
            s2 = self.read_bits(8)
        else:  # aligned data
            s1, s2 = self.read_aligned_bytes(2)
        return (s2 << 8) + s1

    def read_signed_short(self) -> int:
        """Read a signed short (16 bit) from buffer."""
        value = self.read_unsigned_short()
        if value & 0x8000:
            # 2er complement
            return -((~value & 0xFFFF) + 1)
        else:
            return value

    def read_unsigned_long(self) -> int:
        """Read an unsigned long (32 bit) from buffer."""
        if self.bit_index & 7:
            read_bits = self.read_bits
            l1 = read_bits(8)
            l2 = read_bits(8)
            l3 = read_bits(8)
            l4 = read_bits(8)
        else:  # aligned data
            l1, l2, l3, l4 = self.read_aligned_bytes(4)
        return (l4 << 24) + (l3 << 16) + (l2 << 8) + l1

    def read_signed_long(self) -> int:
        """Read a signed long (32 bit) from buffer."""
        value = self.read_unsigned_long()
        if value & 0x80000000:
            # 2er complement
            return -((~value & 0xFFFFFFFF) + 1)
        else:
            return value

    def read_float(self) -> float:
        if self.bit_index & 7:
            read_bits = self.read_bits
            data = bytes(read_bits(8) for _ in range(8))
        else:  # aligned data
            data = bytes(self.read_aligned_bytes(8))
        return struct.unpack("<d", data)[0]

    def read_3_bits(self) -> int:
        bit = self.read_bit()
        if bit:  # 1
            bit = self.read_bit()
            if bit:  # 11
                bit = self.read_bit()
                if bit:
                    return 7  # 111
                else:
                    return 6  # 110
            return 2  # 10
        else:
            return 0  # 0

    @overload
    def read_bit_short(self) -> int:
        ...

    @overload
    def read_bit_short(self, count: int) -> Sequence[int]:
        ...

    def read_bit_short(self, count: int = 1) -> Union[int, Sequence[int]]:
        def _read():
            bits = self.read_bits(2)
            if bits == 0:
                return self.read_signed_short()
            elif bits == 1:
                return self.read_unsigned_byte()
            elif bits == 2:
                return 0
            else:
                return 256

        if count == 1:
            return _read()
        else:
            return tuple(_read() for _ in range(count))

    @overload
    def read_bit_long(self) -> int:
        ...

    @overload
    def read_bit_long(self, count: int) -> Sequence[int]:
        ...

    def read_bit_long(self, count: int = 1) -> Union[int, Sequence[int]]:
        def _read():
            bits = self.read_bits(2)
            if bits == 0:
                return self.read_signed_long()
            elif bits == 1:
                return self.read_unsigned_byte()
            elif bits == 2:
                return 0
            else:  # not used!
                return 256  # ???

        if count == 1:
            return _read()
        else:
            return tuple(_read() for _ in range(count))

    # LibreDWG: https://github.com/LibreDWG/libredwg/blob/master/src/bits.c
    # Read 1 bitlonglong (compacted uint64_t) for REQUIREDVERSIONS, preview_size.
    # ODA doc bug. ODA say 1-3 bits until the first 0 bit. See 3BLL.
    # The first 3 bits indicate the length l (see paragraph 2.1). Then
    # l bytes follow, which represent the number (the least significant
    # byte is first).
    def read_bit_long_long(self) -> int:
        value = 0
        shifting = 0
        length = self.read_bits(3)  # or read_3_bits() ?
        while length > 0:
            value += self.read_unsigned_byte() << shifting
            length -= 1
            shifting += 8
        return value

    @overload
    def read_raw_double(self) -> float:
        ...

    @overload
    def read_raw_double(self, count: int) -> Sequence[float]:
        ...

    def read_raw_double(self, count: int = 1) -> Union[float, Sequence[float]]:
        if count == 1:
            return self.read_float()
        else:
            return tuple(self.read_float() for _ in range(count))

    @overload
    def read_bit_double(self) -> float:
        ...

    @overload
    def read_bit_double(self, count: int) -> Sequence[float]:
        ...

    def read_bit_double(self, count: int = 1) -> Union[float, Sequence[float]]:
        def _read():
            bits = self.read_bits(2)
            if bits == 0:
                return self.read_float()
            elif bits == 1:
                return 1.0
            elif bits == 2:
                return 0.0
            else:  # not used!
                return 0.0

        if count == 1:
            return _read()
        else:
            return tuple(_read() for _ in range(count))

    @overload
    def read_bit_double_default(self) -> float:
        ...

    @overload
    def read_bit_double_default(self, count: int) -> Sequence[float]:
        ...

    @overload
    def read_bit_double_default(
        self, count: int, default: float
    ) -> Sequence[float]:
        ...

    def read_bit_double_default(
        self, count: int = 1, default: float = 0.0
    ) -> Union[float, Sequence[float]]:
        data = struct.pack("<d", default)

        def _read():
            bits = self.read_bits(2)
            if bits == 0:
                return default
            elif bits == 1:
                _data = (
                    bytes(self.read_unsigned_byte() for _ in range(4))
                    + data[4:]
                )
                return struct.unpack("<d", _data)[0]
            elif bits == 2:
                _data = bytearray(data)
                _data[4] = self.read_unsigned_byte()
                _data[5] = self.read_unsigned_byte()
                _data[0] = self.read_unsigned_byte()
                _data[1] = self.read_unsigned_byte()
                _data[2] = self.read_unsigned_byte()
                _data[3] = self.read_unsigned_byte()
                return struct.unpack("<d", _data)[0]
            else:
                return self.read_float()

        if count == 1:
            return _read()
        else:
            return tuple(_read() for _ in range(count))

    def read_signed_modular_chars(self) -> int:
        """Modular characters are a method of storing compressed integer
        values. They consist of a stream of bytes, terminating when the high
        bit (8) of the byte is 0 else another byte follows. Negative numbers
        are indicated by bit 7 set in the last byte.

        """
        shifting = 0
        value = 0
        while True:
            char = self.read_unsigned_byte()
            if char & 0x80:
                # bit 8 set = another char follows
                value |= (char & 0x7F) << shifting
                shifting += 7
            else:
                # bit 8 clear = end of modular char
                # bit 7 set = negative number
                value |= (char & 0x3F) << shifting
                return -value if char & 0x40 else value

    def read_unsigned_modular_chars(self) -> int:
        """Modular characters are a method of storing compressed integer
        values. They consist of a stream of bytes, terminating when the high
        bit (8) of the byte is 0 else another byte follows.

        """
        shifting = 0
        value = 0
        while True:
            char = self.read_unsigned_byte()
            value |= (char & 0x7F) << shifting
            shifting += 7
            # bit 8 set = another char follows
            if not (char & 0x80):
                return value

    def read_modular_shorts(self) -> int:
        """Modular shorts are a method of storing compressed unsigned integer
        values. Only 1 or 2 shorts in practical usage (1GB), if the high
        bit (16) of the first short is set another short follows.

        """
        short = self.read_unsigned_short()
        if short & 0x8000:
            return (self.read_unsigned_short() << 15) | (short & 0x7FFF)
        else:
            return short

    def read_bit_extrusion(self) -> Sequence[float]:
        if self.read_bit():
            return 0.0, 0.0, 1.0
        else:
            return self.read_bit_double(3)

    def read_bit_thickness(self, dxfversion="AC1015") -> float:
        if dxfversion >= "AC1015":
            if self.read_bit():
                return 0.0
        return self.read_bit_double()

    def read_cm_color(self) -> int:
        return self.read_bit_short()

    def read_text(self) -> str:
        length = self.read_bit_short()
        data = bytes(self.read_unsigned_byte() for _ in range(length))
        return data.decode(encoding=self.encoding)

    def read_text_unicode(self) -> str:
        # Unicode text is read from the "string stream" within the object data,
        # see the main Object description section for details.
        length = self.read_bit_short()
        data = bytes(self.read_unsigned_byte() for _ in range(length * 2))
        return data.decode(encoding="utf16")

    def read_text_variable(self) -> str:
        if self.dxfversion < "AC1018":  # R2004
            return self.read_text()
        else:
            return self.read_text_unicode()

    def read_cm_color_cms(self) -> tuple[int, str, str]:
        """Returns tuple (rgb, color_name, book_name)."""
        _ = self.read_bit_short()  # index always 0
        color_name = ""
        book_name = ""
        rgb = self.read_bit_long()
        rc = self.read_unsigned_byte()
        if rc & 1:
            color_name = self.read_text_variable()
        if rc & 2:
            book_name = self.read_text_variable()
        return rgb, color_name, book_name

    def read_cm_color_enc(self) -> Union[int, Sequence[Optional[int]]]:
        """Returns color index as int or tuple (rgb, color_handle,
        transparency_type, transparency).
        """
        flags_and_index = self.read_bit_short()
        flags = flags_and_index >> 8
        index = flags_and_index & 0xFF
        if flags:
            rgb = None
            color_handle = None
            transparency_type = None
            transparency = None
            if flags & 0x80:
                rgb = self.read_bit_short() & 0x00FFFFFF
            if flags & 0x40:
                color_handle = self.read_handle()
            if flags & 0x20:
                data = self.read_bit_long()
                transparency_type = data >> 24
                transparency = data & 0xFF
            return rgb, color_handle, transparency_type, transparency
        else:
            return index

    def read_object_type(self) -> int:
        bits = self.read_bits(2)
        if bits == 0:
            return self.read_unsigned_byte()
        elif bits == 1:
            return self.read_unsigned_byte() + 0x1F0
        else:
            return self.read_unsigned_short()

    def read_handle(self, reference: int = 0) -> int:
        """Returns handle as integer value."""
        code = self.read_bits(4)
        length = self.read_bits(4)
        if code == 6:
            return reference + 1
        if code == 8:
            return reference - 1

        data = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00")
        for index in range(length):
            data[index] = self.read_unsigned_byte()
        offset = struct.unpack("<Q", data)[0]

        if code < 6:
            return offset
        else:
            if code == 10:
                return reference + offset
            if code == 12:
                return reference - offset
        return 0

    def read_hex_handle(self, reference: int = 0) -> str:
        """Returns handle as hex string."""
        return "%X" % self.read_handle(reference)

    def read_code(self, code: str):
        """Read data from bit stream by data codes defined in the
        ODA reference.

        """
        if code == "B":
            return self.read_bit()
        elif code == "RC":
            return self.read_unsigned_byte()
        elif code == "RS":
            return self.read_signed_short()
        elif code == "BS":
            return self.read_bit_short()
        elif code == "RL":
            return self.read_signed_long()
        elif code == "BL":
            return self.read_bit_long()
        elif code == "RD":
            return self.read_raw_double()
        elif code == "2RD":
            return self.read_raw_double(2)
        elif code == "BD":
            return self.read_bit_double()
        elif code == "2BD":
            return self.read_bit_double(2)
        elif code == "3BD":
            return self.read_bit_double(3)
        elif code == "T":
            return self.read_text()
        elif code == "TV":
            return self.read_text_variable()
        elif code == "H":
            return self.read_hex_handle()
        elif code == "BLL":
            return self.read_bit_long_long()
        elif code == "CMC":
            return self.read_cm_color()
        raise ValueError(f"Unknown code: {code}")



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, TYPE_CHECKING, Optional

from ezdxf.lldxf.tags import Tags, group_tags
from ezdxf.math import Vec2, UVec
from ezdxf.tools import pattern

if TYPE_CHECKING:
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["Pattern", "PatternLine"]


class Pattern:
    def __init__(self, lines: Optional[Iterable[PatternLine]] = None):
        self.lines: list[PatternLine] = list(lines) if lines else []

    @classmethod
    def load_tags(cls, tags: Tags) -> Pattern:
        grouped_line_tags = group_tags(tags, splitcode=53)
        return cls(
            PatternLine.load_tags(line_tags) for line_tags in grouped_line_tags
        )

    def clear(self) -> None:
        """Delete all pattern definition lines."""
        self.lines = []

    def add_line(
        self,
        angle: float = 0,
        base_point: UVec = (0, 0),
        offset: UVec = (0, 0),
        dash_length_items: Optional[Iterable[float]] = None,
    ) -> None:
        """Create a new pattern definition line and add the line to the
        :attr:`Pattern.lines` attribute.

        """
        assert (
            dash_length_items is not None
        ), "argument 'dash_length_items' is None"
        self.lines.append(
            PatternLine(angle, base_point, offset, dash_length_items)
        )

    def export_dxf(self, tagwriter: AbstractTagWriter, force=False) -> None:
        if len(self.lines) or force:
            tagwriter.write_tag2(78, len(self.lines))
            for line in self.lines:
                line.export_dxf(tagwriter)

    def __str__(self) -> str:
        return "[" + ",".join(str(line) for line in self.lines) + "]"

    def as_list(self) -> list:
        return [line.as_list() for line in self.lines]

    def scale(self, factor: float = 1, angle: float = 0) -> None:
        """Scale and rotate pattern.

        Be careful, this changes the base pattern definition, maybe better use
        :meth:`Hatch.set_pattern_scale` or :meth:`Hatch.set_pattern_angle`.

        Args:
            factor: scaling factor
            angle: rotation angle in degrees

        """
        scaled_pattern = pattern.scale_pattern(
            self.as_list(), factor=factor, angle=angle
        )
        self.clear()
        for line in scaled_pattern:
            self.add_line(*line)


class PatternLine:
    def __init__(
        self,
        angle: float = 0,
        base_point: UVec = (0, 0),
        offset: UVec = (0, 0),
        dash_length_items: Optional[Iterable[float]] = None,
    ):
        self.angle: float = float(angle)  # in degrees
        self.base_point: Vec2 = Vec2(base_point)
        self.offset: Vec2 = Vec2(offset)
        self.dash_length_items: list[float] = (
            [] if dash_length_items is None else list(dash_length_items)
        )
        # dash_length_items = [item0, item1, ...]
        # item > 0 is line, < 0 is gap, 0.0 = dot;

    @staticmethod
    def load_tags(tags: Tags) -> PatternLine:
        p = {53: 0, 43: 0, 44: 0, 45: 0, 46: 0}
        dash_length_items = []
        for tag in tags:
            code, value = tag
            if code == 49:
                dash_length_items.append(value)
            else:
                p[code] = value
        return PatternLine(
            p[53], (p[43], p[44]), (p[45], p[46]), dash_length_items
        )

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        write_tag = tagwriter.write_tag2
        write_tag(53, self.angle)
        write_tag(43, self.base_point.x)
        write_tag(44, self.base_point.y)
        write_tag(45, self.offset.x)
        write_tag(46, self.offset.y)
        write_tag(79, len(self.dash_length_items))
        for item in self.dash_length_items:
            write_tag(49, item)

    def __str__(self):
        return (
            f"[{self.angle}, {self.base_point}, {self.offset}, "
            f"{self.dash_length_items}]"
        )

    def as_list(self) -> list:
        return [
            self.angle,
            self.base_point,
            self.offset,
            self.dash_length_items,
        ]



# License: MIT License
from ezdxf.addons.pycsg import CSG, Vec3, BSPNode, Polygon
from ezdxf.render.forms import cube, sphere, cone_2p, cylinder_2p


def test_cube_intersect():
    a = cube()
    b = cube().translate(0.5, 0.5)
    c = CSG(mesh=a) * CSG(mesh=b)


def test_cube_union():
    a = cube()
    b = cube().translate(0.5, 0.5)
    c = CSG(mesh=a) + CSG(mesh=b)


def test_cube_subtract():
    a = cube()
    b = cube().translate(0.5, 0.5)
    _ = CSG(a) - CSG(b)


def test_sphere_cylinder_intersect():
    a = sphere(count=8, stacks=4, radius=0.5).translate(0.5, 0.5, 0.5)
    b = cylinder_2p(
        count=16, base_center=(0, 0, 0), top_center=(1, 0, 0), radius=0.3
    )
    CSG(a) * CSG(b)


def test_sphere_cylinder_union():
    a = sphere(count=8, stacks=4, radius=0.5).translate(0.5, 0.5, 0.5)
    b = cylinder_2p(
        count=16, base_center=(0, 0, 0), top_center=(1, 0, 0), radius=0.3
    )
    CSG(a) + CSG(b)


def test_sphere_cylinder_subtract():
    a = sphere(count=8, stacks=4, radius=0.5).translate(0.5, 0.5, 0.5)
    b = cylinder_2p(
        count=16, base_center=(0, 0, 0), top_center=(1, 0, 0), radius=0.3
    )
    CSG(a) - CSG(b)


def test_bolt():
    shaft = cylinder_2p(
        count=32, base_center=(0, 0, 0), top_center=(1, 0, 0), radius=0.1
    )
    head = cone_2p(base_center=(-0.12, 0, 0), apex=(0.1, 0, 0), radius=0.25)
    notch1 = cube().translate(-0.1, 0, 0).scale(0.02, 0.20, 0.02)
    notch2 = cube().translate(-0.1, 0, 0).scale(0.02, 0.02, 0.20)
    bolt = CSG(shaft) + CSG(head) - CSG(notch1) - CSG(notch2)


def test_example_simple():
    v0 = Vec3(0.0, 0.0, 0.0)
    v1 = Vec3(1.0, 0.0, 0.0)
    v2 = Vec3(1.0, 1.0, 0.0)
    p0 = Polygon([v0, v1, v2])
    polygons = [p0]
    node = BSPNode(polygons)


def test_example_infiniteRecursion():
    # This polygon is not exactly planar, causing
    # an infinite recursion when building the BSP
    # tree. Because of the last node, polygon is
    # put at the back of the list with respect to
    # its own cutting plane -- it should be classified
    # as co-planar
    v0 = Vec3(0.12, -0.24, 1.50)
    v1 = Vec3(0.01, 0.00, 1.75)
    v2 = Vec3(-0.03, 0.05, 1.79)
    v3 = Vec3(-0.13, -0.08, 1.5)
    p0 = Polygon([v0, v1, v2, v3])
    polygons = [p0]
    node = BSPNode(polygons)



# Copyright (C) 2011-2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.entities.lwpolyline import (
    format_point,
    compile_array,
    LWPolylinePoints,
    LWPolyline,
)
from ezdxf.math import Vec3


def lwpolyline(points, dxfattribs=None):
    line = LWPolyline.new(dxfattribs=dxfattribs)
    line.set_points(points)
    return line


def lwtags(text):
    tags = ExtendedTags.from_text(text)
    return tags.get_subclass("AcDbPolyline")


def test_new_line():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)

    assert points == list(line.vertices())
    assert 3 == len(line)

    assert line.closed is False, "Polyline should be open by default."
    # test callback DXF attribute
    assert 3 == line.dxf.count  # DXF attrib callback
    with pytest.raises(ezdxf.DXFAttributeError):
        line.dxf.count = 0  # set dxf tag 90


def test_get_point():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    assert (1, 1, 0, 0, 0) == tuple(line[0])
    assert (3, 3, 0, 0, 0) == tuple(line[-1])


def test_slicing():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    assert len(line[:1]) == 1
    assert len(line[:]) == 3
    assert len(line[:2]) == 2
    assert len(line[:-1]) == 2
    assert len(line[1:-1]) == 1
    result = line[::-1]
    assert len(result) == 3
    assert tuple(result[0]) == (3, 3, 0, 0, 0)
    assert tuple(result[1]) == (2, 2, 0, 0, 0)
    assert tuple(result[2]) == (1, 1, 0, 0, 0)


def test_set_point():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)

    line[0] = (4, 4)
    assert (4, 4, 0, 0, 0) == tuple(line[0])

    line[-1] = (4, 4)
    assert (4, 4, 0, 0, 0) == tuple(line[-1])


def test_get_point_error():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    with pytest.raises(IndexError):
        line[3]

def test_insert_point():
    points = [(1, 1), (2, 2), (3, 3), (4, 4), (5, 5)]
    line = lwpolyline(points)
    assert len(line) == 5

    line.insert(0, (7, 8))
    assert len(line) == 6
    assert tuple(line[0]) == (7, 8, 0, 0, 0)
    assert tuple(line[1]) == (1, 1, 0, 0, 0)
    assert tuple(line[-1]) == (5, 5, 0, 0, 0)

    line.insert(1, (1, 9, 4), format="bxy")
    assert len(line) == 7
    assert tuple(line[0]) == (7, 8, 0, 0, 0)
    assert tuple(line[1]) == (9, 4, 0, 0, 1)
    assert tuple(line[2]) == (1, 1, 0, 0, 0)
    assert tuple(line[-1]) == (5, 5, 0, 0, 0)


def test_del_points():
    points = [(1, 1), (2, 2), (3, 3), (4, 4), (5, 5)]
    line = lwpolyline(points)
    assert len(line) == 5

    del line[0]
    assert len(line) == 4
    assert tuple(line[0]) == (2, 2, 0, 0, 0)
    assert tuple(line[-1]) == (5, 5, 0, 0, 0)

    del line[:2]
    assert len(line) == 2
    assert tuple(line[0]) == (4, 4, 0, 0, 0)
    assert tuple(line[-1]) == (5, 5, 0, 0, 0)


def test_append_points():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    line.append_points([(4, 4), (5, 5)])
    assert (4, 4, 0, 0, 0) == tuple(line[-2])


def test_context_manager():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    with line.points() as p:
        p.extend([(4, 4), (5, 5)])
    assert (4, 4, 0, 0, 0) == tuple(line[-2])


def test_clear():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    line.closed = True
    assert line.closed is True, "Polyline should be closed"
    line.clear()
    assert 0 == len(line), "Polyline count should be 0."
    assert (
        len(list(line.get_points())) == 0
    ), "Polyline should not have any points."
    assert line.closed is True, "Polyline should be closed"


def test_delete_const_width():
    points = [(1, 1), (2, 2), (3, 3)]
    line = lwpolyline(points)
    line.dxf.const_width = 0.1
    assert 0.1 == pytest.approx(line.dxf.const_width)
    del line.dxf.const_width


def test_vertices():
    points = [(0, 0, 1, 1, 1), (2, 2, 1, 1, 1), (3, 3, 1, 1, 1)]
    line = lwpolyline(points)
    assert list(line.vertices()) == [(0, 0), (2, 2), (3, 3)]


def test_format_point():
    assert format_point((1, 2, 3, 4, 5), "xy") == (1, 2)
    assert format_point((1, 2, 3, 4, 5), "bse") == (5, 3, 4)
    assert format_point((1, 2, 3, 4, 5), "v,b") == ((1, 2), 5)


def test_point_to_array():
    assert tuple(compile_array((1, 2), "xy")) == (1, 2, 0, 0, 0)
    assert tuple(compile_array((1, 2, 5), "xyb")) == (1, 2, 0, 0, 5)
    assert tuple(compile_array((1, 2), "xyseb")) == (1, 2, 0, 0, 0)

    assert tuple(compile_array((1, 2, 5), "xy")) == (1, 2, 0, 0, 0)
    assert tuple(compile_array((5, (1, 2)), "b,v")) == (1, 2, 0, 0, 5)
    # mix of x, y, v codes is allowed, but only last is set
    assert tuple(compile_array(((1, 2), 4, 5), "vxy")) == (4, 5, 0, 0, 0)


def test_packed_points_basics():
    packed_points, _ = LWPolylinePoints.from_tags(lwtags(LWPOLYLINE1))
    assert len(packed_points) == 2
    points = list(packed_points)
    assert len(points) == 2
    assert tuple(packed_points[0]) == (-0.5, -0.5, 0, 0, 0)
    assert tuple(packed_points[1]) == (0.5, 0.5, 0, 0, 0)
    # test negative index
    assert tuple(packed_points[-1]) == (0.5, 0.5, 0, 0, 0)
    with pytest.raises(IndexError):
        packed_points[-3]
    with pytest.raises(IndexError):
        packed_points[2]


def test_packed_points_advanced():
    packed_points, _ = LWPolylinePoints.from_tags(lwtags(LWPOLYLINE1))
    packed_points.append((5, 5, 1, 2, 3))
    assert len(packed_points) == 3
    assert tuple(packed_points[-1]) == (5, 5, 1, 2, 3)
    packed_points[0] = (7, 7, 0, 0, 0)
    assert tuple(packed_points[0]) == (7, 7, 0, 0, 0)
    assert len(packed_points) == 3
    packed_points.clear()
    assert len(packed_points) == 0


def test_packed_points_to_dxf_tags():
    packed_points, _ = LWPolylinePoints.from_tags(lwtags(LWPOLYLINE1))
    tags = list(packed_points.dxftags())
    assert len(tags) == 2  # just the points
    assert tags[0] == (10, (-0.5, -0.5))
    assert tags[1] == (10, (0.5, 0.5))


def test_packed_points_to_dxf_tags_with_bulge():
    packed_points, _ = LWPolylinePoints.from_tags(lwtags(LWPOLYLINE1))
    packed_points[0] = (-0.5, -0.5, 0, 0, 1)
    packed_points[1] = (0.5, 0.5, 0.1, 0.2, -1)
    tags = list(packed_points.dxftags())
    assert len(tags) == 6
    assert tags[0] == (10, (-0.5, -0.5))
    assert tags[1] == (42, 1)
    assert tags[2] == (10, (0.5, 0.5))
    assert tags[3] == (40, 0.1)
    assert tags[4] == (41, 0.2)
    assert tags[5] == (42, -1)


def test_lwpolyline_transform_interface():
    pline = LWPolyline()
    pline.set_points([(0, 0), (2, 0), (1, 1)], format="xy")
    pline.translate(1, 1, 1)
    vertices = list(pline.vertices())
    assert vertices[0] == (1, 1)
    assert vertices[1] == (3, 1)
    assert vertices[2] == (2, 2)
    assert pline.dxf.elevation == 1
    assert Vec3(0, 0, 1).isclose(pline.dxf.extrusion)


LWPOLYLINE1 = """0
LWPOLYLINE
5
239
330
238
100
AcDbEntity
8
0
6
ByBlock
62
0
100
AcDbPolyline
90
2
70
0
43
0.15
10
-0.5
20
-0.5
10
0.5
20
0.5
"""



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ezdxf.document import Drawing


class Messenger:
    def __init__(self, doc: Drawing) -> None:
        self.doc = doc
        self.entitydb = doc.entitydb

    def broadcast(self, message_type: int, data: Any = None) -> None:
        """Broadcast a message to all entities in the entity database."""
        # Receiver can change the entity database while processing the message.
        for entity in list(self.entitydb.values()):
            entity.notify(message_type, data)
